import json
import os.path
import sys
import typing as t

import click
import sentry_sdk
from click import Context, Parameter
from rich.console import Console

from piperider_cli import __version__, event, sentry_dns, sentry_env
from piperider_cli.cli_utils import DbtUtil
from piperider_cli.cli_utils.cloud import CloudConnectorHelper
from piperider_cli.configuration import FileSystem, is_piperider_workspace_exist
from piperider_cli.error import DbtProjectNotFoundError
from piperider_cli.event import UserProfileConfigurator
from piperider_cli.event.track import TrackCommand
from piperider_cli.feedback import Feedback
from piperider_cli.generate_report import GenerateReport
from piperider_cli.guide import Guide
from piperider_cli.initializer import Initializer

release_version = __version__ if sentry_env != 'development' else None

sentry_sdk.init(
    sentry_dns,
    environment=sentry_env,
    release=release_version,
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for performance monitoring.
    # We recommend adjusting this value in production.
    traces_sample_rate=1.0
)
sentry_sdk.set_tag("piperider.version", __version__)
sentry_sdk.set_tag("platform", sys.platform)

event.init()

debug_option = [
    click.option('--debug', is_flag=True, help='Enable debug mode.')
]


def dbt_select_option_builder():
    try:
        # For dbt-core >= 1.5.0
        from dbt.cli.options import MultiOption
        return click.option('--select', '-s', default=None, type=tuple, help='Specify the dbt nodes to include.',
                            multiple=True,
                            cls=MultiOption)
    except Exception:
        # For dbt-core < 1.5.0
        return click.option('--select', '-s', default=None, help='Specify the dbt nodes to include.',
                            multiple=True)


dbt_related_options = [
    click.option('--dbt-project-dir', type=click.Path(exists=True),
                 help='The path to the dbt project directory.'),
    click.option('--dbt-profiles-dir', type=click.Path(exists=True), default=None,
                 help='Directory to search for dbt profiles.yml.'),
    click.option('--no-auto-search', type=click.BOOL, default=False, is_flag=True,
                 help='Disable auto detection of dbt projects.'),
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


class RunDataPath(click.Path):
    def __init__(self):
        super().__init__(exists=True, dir_okay=False, resolve_path=True)

    def convert(
            self, value: t.Any, param: t.Optional["Parameter"], ctx: t.Optional["Context"]
    ) -> t.Any:
        rv = value

        try:
            data = self.read_file(rv)
            self.validate(data)
        except BaseException as e:
            self.fail(f'{e} from: "{value}"')
            pass

        return super().convert(rv, param, ctx)

    def read_file(self, rv):
        try:
            with open(rv) as fh:
                return json.loads(fh.read())
        except BaseException:
            raise ValueError('failed to read data')

    def validate(self, data):
        if not isinstance(data, t.Dict):
            raise ValueError('type mismatched, we expected a dict')
        required_keys = {'tables', 'datasource'}
        if not set(data.keys()).issuperset(required_keys):
            raise ValueError(f'missing one of required keys: {required_keys}')

        if 'dbt' in data and data.get('dbt', {}).get('manifest') is None:
            raise ValueError('there is no dbt manifest data in the run.json')


@click.group(name="piperider", invoke_without_command=True)
@click.pass_context
def cli(ctx: click.Context):
    'Code review for data in dbt'
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        Guide().show_tips(ctx.command.name)


# It works after click>=8
cli.command_class = TrackCommand


@cli.command(short_help='Show version information.', cls=TrackCommand)
def version():
    'Show version information.'
    console = Console()
    console.print(__version__)

    if 'production' == sentry_env:
        return

    from piperider_cli import data
    commit_file = os.path.abspath(os.path.join(os.path.dirname(data.__file__), 'COMMIT'))
    try:
        with open(commit_file) as fh:
            commit_sha = fh.read().strip()
            console.print(f'GitCommit: {commit_sha}')
    except Exception:
        pass


@cli.command(short_help='Send your feedback to help us improve the PipeRider.', cls=TrackCommand)
def feedback():
    'Send your feedback to help us improve the PipeRider.'
    Feedback.exec()


@cli.command(short_help='Initialize a PipeRider project.', cls=TrackCommand)
@add_options(dbt_related_options)
@add_options(debug_option)
def init(**kwargs):
    'Initialize a PipeRider project in interactive mode. The configurations are saved in ".piperider".'

    console = Console()

    # Search dbt project config files
    dbt_project_dir = kwargs.get('dbt_project_dir')
    no_auto_search = kwargs.get('no_auto_search')
    dbt_project_path = DbtUtil.get_dbt_project_path(dbt_project_dir, no_auto_search)
    dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
    if dbt_project_path:
        FileSystem.set_working_directory(dbt_project_path)

    # TODO show the process and message to users
    console.print(f'Initialize piperider to path {FileSystem.PIPERIDER_WORKSPACE_PATH}')

    config = Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir)
    if kwargs.get('debug'):
        for ds in config.dataSources:
            console.rule('Configuration')
            console.print(ds.__dict__)
    if config is None:
        sys.exit(1)

    # Show the content of config.yml
    Initializer.show_config()


@cli.command(short_help='Check the configuraion and connection.', cls=TrackCommand)
@add_options(dbt_related_options)
@add_options(debug_option)
def diagnose(**kwargs):
    'Check project configuration, datasource, connections, and assertion configuration.'

    console = Console()

    # Search dbt project config files
    dbt_project_dir = kwargs.get('dbt_project_dir')
    no_auto_search = kwargs.get('no_auto_search')
    dbt_project_path = DbtUtil.get_dbt_project_path(dbt_project_dir, no_auto_search)
    dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
    if dbt_project_path:
        FileSystem.set_working_directory(dbt_project_path)
        # Only run initializer when dbt project path is provided
        Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir, interactive=False)
    elif is_piperider_workspace_exist() is False:
        raise DbtProjectNotFoundError()

    console.print('Diagnosing...')

    console.print(f'[bold dark_orange]PipeRider Version:[/bold dark_orange] {__version__}')

    from piperider_cli.validator import Validator
    if not Validator.diagnose():
        sys.exit(1)


@cli.command(short_help='Execute a piperider run.', cls=TrackCommand)
@click.option('--datasource', default=None, type=click.STRING, help='Datasource to use.', metavar='DATASOURCE_NAME')
@click.option('--table', default=None, type=click.STRING, help='Specify the name of a model/seed/source to profile.',
              metavar='TABLE_NAME')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--skip-report', is_flag=True, help='Skip generating report.')
@click.option('--dbt-target-path', default=None, help='Configure the "target-path" of dbt.')
@click.option('--dbt-list', is_flag=True, help='Associate with dbt list format input.')
@click.option('--dbt-run-results', is_flag=True, help='Associate with dbt run results.',
              hidden=True)  # For backward compatibility
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--upload', is_flag=True, help='Upload the report to the PipeRider Cloud.')
@click.option('--project', default=None, type=click.STRING, help='Specify the project name to upload.')
@click.option('--share', default=False, is_flag=True, help='Enable public share of the report to PipeRider Cloud.')
@click.option('--open', is_flag=True, help='Opens the generated report in the system\'s default browser.')
@click.option('--skip-datasource', default=False, is_flag=True,
              help='Skip accessing the connection of datasource.')
@add_options([
    dbt_select_option_builder(),
    click.option('--state', default=None,
                 help='If set, use the given directory as the source for JSON files to compare with this project.')
])
@add_options(dbt_related_options)
@add_options(debug_option)
def run(**kwargs):
    """
    Profile data source, run assertions, and generate report(s). By default, the raw results and reports are saved in ".piperider/outputs".
    """

    from piperider_cli.cli_utils.run_cmd import run as cmd
    return cmd(**kwargs)


@cli.command(short_help='Generate recommended assertions. - Deprecated', cls=TrackCommand)
@click.option('--input', default=None, type=click.Path(exists=True), help='Specify the raw result file.')
@click.option('--no-recommend', is_flag=True, help='Generate assertions templates only.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--table', default=None, type=click.STRING, help='Generate assertions for the given table')
@add_options(debug_option)
def generate_assertions(**kwargs):
    'Generate recommended assertions based on the latest result. By default, the profiling result will be loaded from ".piperider/outputs".'
    input_path = kwargs.get('input')
    report_dir = kwargs.get('report_dir')
    no_recommend = kwargs.get('no_recommend')
    table = kwargs.get('table')

    from piperider_cli.assertion_generator import AssertionGenerator
    ret = AssertionGenerator.exec(input_path=input_path, report_dir=report_dir, no_recommend=no_recommend, table=table)
    if ret != 0:
        sys.exit(ret)
    return ret


@cli.command(short_help='Generate a report.', cls=TrackCommand)
@click.option('--input', default=None, type=RunDataPath(), help='Specify the raw result file.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@add_options(debug_option)
def generate_report(**kwargs):
    'Generate a report from the latest raw result or specified result. By default, the raw results are saved in ".piperider/outputs".'
    GenerateReport.exec(input=kwargs.get('input'), report_dir=kwargs.get('report_dir'), output=kwargs.get('output'))


@cli.command(short_help='Compare two existing reports.', cls=TrackCommand)
@click.option('--base', default=None, type=RunDataPath(), help='Specify the base report file.')
@click.option('--target', default=None, type=RunDataPath(), help='Specify the report file to be compared.')
@click.option('--last', default=None, is_flag=True, help='Compare the last two reports.')
@click.option('--datasource', default=None, type=click.STRING, metavar='DATASOURCE_NAME',
              help='Specify the datasource.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--summary-file', default=None, type=click.STRING, help='Output the comparison summary markdown file.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--tables-from', default='all',
              type=click.Choice(['all', 'target-only', 'base-only'], case_sensitive=False),
              help='Show table comparison from base or target.')
@click.option('--upload', default=False, is_flag=True, help='Upload the report to PipeRider Cloud.')
@click.option('--project', default=None, type=click.STRING,
              help='Specify the project name to upload.')
@click.option('--share', default=False, is_flag=True, help='Enable public share of the report to PipeRider Cloud.')
@click.option('--open', is_flag=True, help='Opens the generated report in the system\'s default browser')
@add_options(debug_option)
def compare_reports(**kwargs):
    'Compare two existing reports selected in interactive mode or by option.'

    a = kwargs.get('base')
    b = kwargs.get('target')
    last = kwargs.get('last')
    datasource = kwargs.get('datasource')
    tables_from = kwargs.get('tables_from')
    summary_file = kwargs.get('summary_file')
    open_report = kwargs.get('open')
    force_upload = kwargs.get('upload')
    enable_share = kwargs.get('share')
    project_name = kwargs.get('project')

    if enable_share or CloudConnectorHelper.is_auto_upload():
        force_upload = True

    if force_upload and not CloudConnectorHelper.is_login():
        force_upload = False
        console = Console()
        console.print('[bold yellow]Warning: [/bold yellow]Reports will not be uploaded due to not logged in.')

    from piperider_cli.compare_report import CompareReport
    CompareReport.exec(a=a, b=b, last=last, datasource=datasource,
                       report_dir=kwargs.get('report_dir'), output=kwargs.get('output'), summary_file=summary_file,
                       tables_from=tables_from, force_upload=force_upload, enable_share=enable_share,
                       open_report=open_report, project_name=project_name, debug=kwargs.get('debug', False),
                       show_progress=True)


@cli.group('config', short_help='Manage the PipeRider configurations.')
def config(**kwargs):
    pass


@config.command(name='list-datasource', short_help='List DataSources in current PipeRider project',
                cls=TrackCommand)
@add_options(debug_option)
def list(**kwargs):
    'List PipeRider current configurations.'
    Initializer.list()


@config.command(name='add-datasource', short_help='Add a DataSource to current PipeRider project.',
                cls=TrackCommand)
@add_options(debug_option)
def add(**kwargs):
    Initializer.add()
    pass


@config.command(name='delete-datasource', short_help='Delete a DataSource from current PipeRider project.',
                cls=TrackCommand)
@add_options(debug_option)
def delete(**kwargs):
    Initializer.list()
    Initializer.delete()
    pass


@config.command(name='enable-auto-upload', short_help='Enable auto upload to PipeRider Cloud.', cls=TrackCommand)
def enable_auto_upload(**kwargs):
    CloudConnectorHelper.config_auto_upload(True)
    pass


@config.command(name='disable-auto-upload', short_help='Disable auto upload to PipeRider Cloud.', cls=TrackCommand)
def disable_auto_upload(**kwargs):
    CloudConnectorHelper.config_auto_upload(False)
    pass


@config.command(name='enable-user-tracking', short_help='Enable user tracking.', cls=TrackCommand)
def enable_user_tracking(**kwargs):
    UserProfileConfigurator.update('anonymous_tracking', True, name='user tracking')


@config.command(name='disable-user-tracking', short_help='Disable user tracking.', cls=TrackCommand)
def disable_user_tracking(**kwargs):
    UserProfileConfigurator.update('anonymous_tracking', False, name='user tracking')


@cli.group('cloud', short_help='Manage PipeRider Cloud')
def cloud(**kwargs):
    # Manage PipeRider Cloud.
    pass


@cloud.command(short_help='Upload a report to the PipeRider Cloud.', cls=TrackCommand)
@click.option('--run', type=RunDataPath(), help='Specify the raw result file.')
@click.option('--report-dir', default=None, type=click.STRING, help='Use a different report directory.')
@click.option('--datasource', default=None, type=click.STRING, metavar='DATASOURCE_NAME',
              help='Specify the datasource.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project name to upload.')
@add_options(debug_option)
def upload_report(**kwargs):
    """
    Upload a single run report to PipeRider Cloud
    """
    report_path = kwargs.get('run')
    datasource = kwargs.get('datasource')
    report_dir = kwargs.get('report_dir')
    project_name = kwargs.get('project')
    ret = CloudConnectorHelper.upload_report(report_path=report_path, datasource=datasource, report_dir=report_dir,
                                             project_name=project_name,
                                             debug=kwargs.get('debug', False))
    return ret


@cloud.command(name='compare-reports', short_help='Generate comparison report on PipeRider Cloud.', cls=TrackCommand)
@click.option('--base', default=None, required=False, type=click.STRING,
              help='Specify the base report id or data source name. e.g., 123 or datasource:<name>')
@click.option('--target', default=None, required=False, type=click.STRING,
              help='Specify the target report id or data source name. e.g., 123 or datasource:<name>')
@click.option('--tables-from', default='all',
              type=click.Choice(['all', 'target-only', 'base-only'], case_sensitive=False),
              help='Show table comparison from base or target.')
@click.option('--summary-file', default=None, type=click.STRING, help='Download the comparison summary markdown file.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project name to upload.')
@add_options(debug_option)
def cloud_compare_reports(**kwargs):
    """
    Generate comparison report on PipeRider Cloud
    """
    base = kwargs.get('base')
    target = kwargs.get('target')
    tables_from = kwargs.get('tables_from')
    summary_file = kwargs.get('summary_file')
    project_name = kwargs.get('project')

    ret = CloudConnectorHelper.compare_reports(base=base, target=target, tables_from=tables_from,
                                               summary_file=summary_file,
                                               project_name=project_name, debug=kwargs.get('debug', False))

    if ret != 0:
        sys.exit(ret)
    return ret


@cli.command(name='compare', short_help='Compare the change for the current branch.', cls=TrackCommand)
@click.option('--recipe', default=None, type=click.STRING, help='Select a different recipe.')
@click.option('--upload', default=False, is_flag=True, help='Upload the report to PipeRider Cloud.')
@click.option('--share', default=False, is_flag=True, help='Enable public share of the report to PipeRider Cloud.')
@click.option('--output', '-o', default=None, type=click.STRING, help='Directory to save the results.')
@click.option('--summary-file', default=None, type=click.STRING, help='Output the comparison summary markdown file.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the default project name.')
@click.option('--open', is_flag=True, help='Opens the generated report in the system\'s default browser')
@click.option('--dry-run', is_flag=True, default=False, help='Display the run details without actually executing it')
@click.option('--interactive', is_flag=True, default=False,
              help='Prompt for confirmation to proceed with the run (Y/N)')
@add_options([
    dbt_select_option_builder(),
    click.option('--modified', default=False, is_flag=True, help='Only compare the modified models.'),
    click.option('--base-branch', default='main', type=click.STRING, help='Specify the base branch.',
                 show_default=True),
])
@add_options(dbt_related_options)
@add_options(debug_option)
def compare_with_recipe(**kwargs):
    """
    Generate the comparison report for your branch.
    """

    from piperider_cli.cli_utils.compare_with_recipe import compare_with_recipe as cmd
    return cmd(**kwargs)


@cloud.command(short_help='Signup to PipeRider Cloud.', cls=TrackCommand)
@add_options(debug_option)
def signup(**kwargs):
    ret = CloudConnectorHelper.signup(debug=kwargs.get('debug', False))
    return ret


@cloud.command(short_help='Login to PipeRider Cloud.', cls=TrackCommand)
@click.option('--token', default=None, type=click.STRING, help='Specify the API token.')
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the default project name.')
@click.option('--no-interaction', default=False, is_flag=True, help='Disable interactive mode.')
@add_options(debug_option)
def login(**kwargs):
    options = {
        'no_interaction': False,
    }

    if kwargs.get('project') is not None:
        options['default_project'] = kwargs.get('project')
    if kwargs.get('no_interaction') is True:
        options['no_interaction'] = True

    ret = CloudConnectorHelper.login(api_token=kwargs.get('token'), options=options, debug=kwargs.get('debug', False))
    return ret


@cloud.command(short_help='Logout from PipeRider Cloud.', cls=TrackCommand)
@add_options(debug_option)
def logout(**kwargs):
    ret = CloudConnectorHelper.logout()
    return ret


@cloud.command(short_help='List projects on PipeRider Cloud.', cls=TrackCommand)
@add_options(debug_option)
def list_projects(**kwargs):
    ret = CloudConnectorHelper.list_projects()
    return ret


@cloud.command(short_help='Select a project on PipeRider Cloud as default project.', cls=TrackCommand)
@click.option('--project', default=None, type=click.STRING, metavar='PROJECT_NAME',
              help='Specify the project name.')
@click.option('--no-interaction', default=False, is_flag=True, help='Disable interactive mode.')
@add_options(debug_option)
def select_project(**kwargs):
    project_name = kwargs.get('project')
    no_interaction: bool = kwargs.get('no_interaction', False)
    ret = CloudConnectorHelper.select_project(project_name=project_name, no_interaction=no_interaction)
    return ret



import json
import os
import shutil

from piperider_cli.utils import create_link


def test_create_link():
    dirname = os.path.join(os.getcwd(), "__create_link__")
    dirname_latest = os.path.join(os.getcwd(), "__create_link__latest")

    def cleanup():
        if os.path.exists(dirname_latest):
            os.unlink(dirname_latest)
        if os.path.exists(dirname):
            shutil.rmtree(dirname, ignore_errors=True)

    cleanup()

    try:
        os.makedirs(dirname, exist_ok=True)
        create_link(dirname, dirname_latest)
    finally:
        cleanup()



import json
import os
import tempfile
from unittest import TestCase

from sqlalchemy import create_engine, MetaData, Table, Column, Integer

from piperider_cli.assertion_engine import AssertionEngine


def build_test_assertions(assertion_config_text: str):
    assertions_dir = None
    with tempfile.NamedTemporaryFile() as tmp:
        assertions_dir = os.path.join(tmp.name, 'assertions')
    os.makedirs(assertions_dir, exist_ok=True)

    with open(os.path.join(assertions_dir, "assertions.yml"), "w") as fh:
        fh.write(assertion_config_text)
    return assertions_dir


def build_assertion_engine(table, profiler_result, assertions):
    db_engine = create_engine('sqlite://')
    metadata = MetaData()
    Table(table, metadata, Column('misc', Integer))
    metadata.create_all(db_engine)

    engine = AssertionEngine(db_engine, build_test_assertions(assertions))
    engine.load_assertions(profiler_result=profiler_result, config_path=None)
    return engine


class BuiltinAssertionsTests(TestCase):

    def setUp(self) -> None:
        # load metrics
        metrics_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "orders_1k.json")
        with open(metrics_file) as fh:
            self.metrics = json.loads(fh.read())

    def test_assert_row_count_in_range2(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          tests:
          - name: assert_row_count_in_range
            assert:
              count: [1000, 200000.5]
            tags:
            - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)
        results = engine.validate_assertions()

        self.assertEqual(1, len(results))

        self.assertEqual('Found assertion syntax problem => name: assert_row_count_in_range for table '
                         'orders_1k\n'
                         "ERROR: count parameter should be one of the types {<class 'int'>}, input: "
                         '[1000, 200000.5]',
                         results[0].as_internal_report())

    def test_assert_row_count_in_range(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          tests:
          - name: assert_row_count_in_range
            assert:
              count: [1000, 200000]
            tags:
            - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        assertion_result = results[0].result
        self.assertEqual(dict(success=True, exceptions=None),
                         dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_type(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          columns:
            o_totalprice:
              tests:
              - name: assert_column_type
                assert:
                  type: numeric
                tags:
                - OPTIONAL
            o_orderdate:
              tests:
              - name: assert_column_type
                assert:
                  type: datetime
                tags:
                - OPTIONAL
            o_orderpriority:
              tests:
              - name: assert_column_type
                assert:
                  type: string
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        for result in results:
            assertion_result = result.result
            self.assertEqual(dict(success=True, exceptions=None),
                             dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_in_range(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          columns:
            o_totalprice:
              tests:
              - name: assert_column_min_in_range
                assert:
                  min: [1000, 1200]
                tags:
                - OPTIONAL
              - name: assert_column_max_in_range
                assert:
                  max: [440000, 450000]
                tags:
                - OPTIONAL
              - name: assert_column_in_range
                assert:
                  range: [0, 450000]
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        for result in results:
            assertion_result = result.result
            self.assertEqual(dict(success=True, exceptions=None),
                             dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_not_null(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          columns:
            o_totalprice:
              tests:
              - name: assert_column_not_null
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        for result in results:
            assertion_result = result.result
            self.assertEqual(dict(success=True, exceptions=None),
                             dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_unique(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          columns:
            o_totalprice:
              tests:
              - name: assert_column_unique
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        for result in results:
            assertion_result = result.result
            self.assertEqual(dict(success=True, exceptions=None),
                             dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_exist(self):
        assertions = """
        orders_1k:  # Table Name
          # Test Cases for Table
          columns:
            o_totalprice:
              tests:
              - name: assert_column_exist
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)

        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        for result in results:
            assertion_result = result.result
            self.assertEqual(dict(success=True, exceptions=None),
                             dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_value_range(self):
        assertions = """
        orders_1k:  # Table Name
          columns:
            o_totalprice:
              tests:
              - name: assert_column_value
                assert:
                  lte: 440269.51
                  gte: 1106.99
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)
        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        assertion_result = results[0].result
        self.assertEqual(dict(success=True, exceptions=None),
                         dict(success=assertion_result._success, exceptions=assertion_result._exception))

    def test_assert_column_value_set(self):
        assertions = """
        orders_1k:  # Table Name
          columns:
            o_orderstatus:
              tests:
              - name: assert_column_value
                assert:
                  in: ['O', 'F', 'P']
                tags:
                - OPTIONAL
        """
        engine = build_assertion_engine('orders_1k', self.metrics, assertions)
        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        assertion_result = results[0].result
        self.assertEqual(dict(success=True, exceptions=None),
                         dict(success=assertion_result._success, exceptions=assertion_result._exception))



from collections import deque


def topological_sort(graph, num_of_vertices):
    in_degree = {i: 0 for i in graph}

    for i in graph:
        for j in graph[i]:
            in_degree[j] += 1

    queue = deque()
    for i in graph:
        if in_degree[i] == 0:
            queue.append(i)

    cnt = 0
    top_order = []

    while queue:
        u = queue.popleft()
        top_order.append(u)

        for i in graph[u]:
            in_degree[i] -= 1
            if in_degree[i] == 0:
                queue.append(i)

        cnt += 1

    # Check if there was a cycle
    if cnt != num_of_vertices:
        raise Exception("There exists a cycle in the graph. Topological sorting is not possible.")
    else:
        return top_order



import os
import platform
import subprocess


def remove_link(link_path):
    try:
        if platform.system() == 'Windows':
            if os.path.exists(link_path):
                subprocess.run(["rmdir", link_path], shell=True, check=True)
        else:
            if os.path.exists(link_path) or os.path.islink(link_path):
                os.unlink(link_path)
    except OSError as e:
        raise e


def create_link(source: str, target: str):
    if not os.path.isabs(source):
        raise ValueError(f'source must be abspath: {source}')
    if not os.path.isabs(target):
        raise ValueError(f'target must be abspath: {target}')

    remove_link(target)

    # Check the platform
    if platform.system() == "Windows":
        # On Windows, try creating a junction
        try:
            subprocess.run(["mklink", "/J", target, source], shell=True, check=True,
                           stderr=subprocess.DEVNULL,
                           stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            raise e
    else:
        # On non-Windows platforms, create a symbolic link
        os.symlink(source, target)



from sqlalchemy.engine import URL

from piperider_cli.error import PipeRiderConnectorError
from . import DataSource
from .field import TextField


class AthenaDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'athena', **kwargs)
        self.fields = [
            TextField('s3_staging_dir',
                      description='S3 location to store Athena query results and metadata, e.g. s3://athena_query_result/prefix/'),
            TextField('region_name', description='AWS region name of your Athena instance'),
            TextField('database', description='Athena data catalog', default='awsdatacatalog'),
            TextField('schema', description='Athena database'),

        ]

    def validate(self):
        if self.type_name != 'athena':
            raise ValueError('type name should be athena')
        return self._validate_required_fields()

    def to_database_url(self, database):
        credential = self.credential
        catalog_name = credential.get('database')
        schema = credential.get('schema')
        region_name = credential.get('region_name')
        s3_staging_dir = credential.get('s3_staging_dir')

        return f"awsathena+rest://athena.{region_name}.amazonaws.com:443/{schema}?s3_staging_dir={s3_staging_dir}&catalog_name={catalog_name}"

    def verify_connector(self):
        try:
            import pyathena
            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'athena')

    def _get_display_description(self):
        cred = self.credential
        return f"type={self.type_name}, database={cred.get('database')}, schema={cred.get('schema')}"

    def get_database(self):
        return self.credential.get('database')

    def get_schema(self):
        return self.credential.get('schema')



class CloudConnectorHelper:
    @staticmethod
    def is_login() -> bool:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.is_login()

    @staticmethod
    def is_auto_upload() -> bool:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.is_auto_upload()

    @staticmethod
    def config_auto_upload(flag: bool):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.config_auto_upload(flag)

    @staticmethod
    def signup(debug=False):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.signup(debug=debug)

    @staticmethod
    def login(api_token=None, options: dict = None, debug=False):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.login(api_token=api_token, options=options, debug=debug)

    @staticmethod
    def logout():
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.logout()

    @staticmethod
    def upload_latest_report(report_dir=None, debug=False, open_report=False, enable_share=False,
                             project_name: str = None) -> int:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.upload_latest_report(report_dir=report_dir, debug=debug, open_report=open_report,
                                      enable_share=enable_share, project_name=project_name)

    @staticmethod
    def upload_report(report_path=None, report_dir=None, datasource=None, debug=False, open_report=False,
                      enable_share=False, project_name=None, show_progress=True) -> int:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.upload_report(report_path=report_path, report_dir=report_dir, datasource=datasource,
                               debug=debug, open_report=open_report, enable_share=enable_share,
                               project_name=project_name, show_progress=show_progress)

    @staticmethod
    def generate_compare_report(base_id: str, target_id: str, tables_from='all',
                                project_name: str = None, debug: bool = False):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.generate_compare_report(base_id=base_id, target_id=target_id, tables_from=tables_from,
                                         project_name=project_name, debug=debug)

    @staticmethod
    def compare_reports(base=None, target=None, tables_from='all', summary_file=None, debug=False,
                        project_name=None) -> int:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.compare_reports(base=base, target=target, tables_from=tables_from, summary_file=summary_file,
                                 debug=debug, project_name=project_name)

    @staticmethod
    def share_run_report(run_id=None, debug=False, project_name=None):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.share_run_report(run_id=run_id, debug=debug, project_name=project_name)

    @staticmethod
    def share_compare_report(base_id=None, target_id=None, debug=False, project_name=None):
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.share_compare_report(base_id=base_id, target_id=target_id, debug=debug, project_name=project_name)

    @staticmethod
    def list_projects(debug=False) -> int:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.list_projects(debug=debug)

    @staticmethod
    def select_project(project_name: str = None,
                       datasource: str = None,
                       no_interaction: bool = False,
                       debug: bool = False) -> int:
        from piperider_cli.cloud_connector import CloudConnector as c
        return c.select_project(project_name=project_name, datasource=datasource,
                                no_interaction=no_interaction, debug=debug)



import io
import json
import os
import sys
from glob import glob
from pathlib import Path
from typing import Dict, Optional, Union

import inquirer
from rich.console import Console
from rich.table import Table
from ruamel import yaml

from piperider_cli import load_jinja_template, load_jinja_string_template
from piperider_cli.dbt.list_task import load_manifest, list_resources_unique_id_from_manifest, load_full_manifest
from piperider_cli.error import \
    DbtProjectInvalidError, \
    DbtProfileInvalidError, \
    DbtProfileBigQueryAuthWithTokenUnsupportedError, DbtRunTimeError
from piperider_cli.metrics_engine import Metric
from piperider_cli.metrics_engine.metrics import SemanticModel
from piperider_cli.statistics import Statistics

console = Console()


def search_dbt_project_path() -> str:
    paths = list(Path.cwd().parents)
    paths.insert(0, Path.cwd())
    return next((str(x) for x in paths if (x / "dbt_project.yml").exists()), None)


def get_dbt_project_path(dbt_project_dir: str = None, no_auto_search: bool = False,
                         recursive: bool = True) -> str:
    project_dir = dbt_project_dir if dbt_project_dir else search_dbt_project_path()
    if project_dir is None:
        return None
    return os.path.join(project_dir, "dbt_project.yml")


# Deprecated due to we will search parent directory for dbt_project.yml, not subdirectory ref: SC-31557
def _search_dbt_project_path(recursive: bool = True):
    exclude_patterns = ['site-packages', 'dbt_packages']
    segments = [os.getcwd()]
    if recursive:
        _warning_if_search_path_too_widely(os.getcwd())
        segments.append('**')
    paths = glob(os.path.join(*segments, 'dbt_project.yml'), recursive=True)
    for exclude_pattern in exclude_patterns:
        paths = list(filter(lambda x: exclude_pattern not in x, paths))
    dbt_project_path = None
    if not paths:
        # No dbt project found
        return dbt_project_path

    if len(paths) == 1:
        # Only one dbt project found, use it
        dbt_project_path = paths[0]
    else:
        # Multiple dbt projects found, ask user to select one
        paths = sorted(paths, key=len)
        default_index = 0
        console.print(f'\nMultiple dbt projects found. Please select the dbt project: (Default: {default_index})')
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column('Index', style="bold", width=5, justify='right')
        table.add_column('Path', style="bold")
        for i, p in enumerate(paths):
            table.add_row(str(i), p)
        console.print(table)
        in_index = input('index : ').strip()
        try:
            in_index = int(in_index) if in_index else default_index
            index = int(in_index)
            dbt_project_path = paths[index]
        except Exception as e:
            console.print(f'[bold yellow]Failed to read index![/] Error: {e}')
            return None

    return dbt_project_path


def _warning_if_search_path_too_widely(search_path):
    # Only warning user on macOS platform
    if sys.platform != "darwin":
        return

    home_dir = os.path.expanduser('~')

    if search_path in home_dir:
        console.print(
            f"[[bold yellow]Warning[/bold yellow]] Search path '{search_path}' is too widely. It will take some time to parse directories and may need extra permissions.")
        if inquirer.confirm(message="Do you still want to keep going?", default=True) is not True:
            raise KeyboardInterrupt()


def is_ready(config):
    if not config.get('profile'):
        return False

    if not config.get('target'):
        return False

    if not config.get('projectDir'):
        return False

    if not os.path.expanduser(config.get('projectDir')):
        return False

    return True


def _is_dbt_file_existing(dbt_state_dir: str, filename: str):
    dbt_state_dir = os.path.abspath(dbt_state_dir)
    filepath = os.path.join(dbt_state_dir, filename)
    if not os.path.exists(filepath):
        return False
    return True


def is_dbt_state_ready(dbt_state_dir: str):
    return _is_dbt_file_existing(dbt_state_dir, 'manifest.json')


def is_dbt_run_results_ready(dbt_state_dir: str):
    return _is_dbt_file_existing(dbt_state_dir, 'run_results.json')


def _get_state_run_results(dbt_state_dir: str):
    path = os.path.join(dbt_state_dir, 'run_results.json')
    with open(path) as f:
        run_results = json.load(f)

    return run_results


def _get_state_manifest(dbt_state_dir: str):
    path = os.path.join(dbt_state_dir, 'manifest.json')
    if os.path.isabs(path) is False:
        from piperider_cli.configuration import FileSystem
        path = os.path.join(FileSystem.WORKING_DIRECTORY, path)
    with open(path) as f:
        manifest = json.load(f)

    return manifest


def append_descriptions(profile_result, dbt_state_dir):
    manifest = _get_state_manifest(dbt_state_dir)

    nodes = manifest.get('nodes')
    for node in nodes.values():
        model = node.get('name')
        model_desc = node.get('description')
        if model not in profile_result['tables']:
            continue
        if model_desc:
            profile_result['tables'][model]['description'] = f"{model_desc}"

        columns = node.get('columns', {})
        for column, v in columns.items():
            if column not in profile_result['tables'][model]['columns']:
                continue
            column_desc = v.get('description')
            if column_desc:
                profile_result['tables'][model]['columns'][column]['description'] = f"{column_desc}"


def get_dbt_state_candidate(dbt_state_dir: str, options: dict, *, select_for_metadata: bool = False):
    candidate = []
    material_whitelist = ['seed', 'table', 'incremental']
    resource_whitelist = ['model', 'seed']
    if options.get('view_profile'):
        material_whitelist.append('view')

    tag = options.get('tag')
    dbt_resources = options.get('dbt_resources')

    manifest = _get_state_manifest(dbt_state_dir)
    nodes = manifest.get('nodes')
    sources = manifest.get('sources')

    def profiling_chosen_fn(key, node):
        statistics = Statistics()
        if dbt_resources:
            chosen = node.get('unique_id') in dbt_resources['models']
            if not chosen:
                statistics.add_field_one('filter')
            return chosen
        else:
            if tag:
                chosen = tag in node.get('tags', [])
                if not chosen:
                    statistics.add_field_one('notag')
                return chosen
            config_material = node.get('config').get('materialized')
            if config_material not in material_whitelist:
                statistics.add_field_one(config_material)
                return False
            return True

    def metadata_chosen_fn(key, node):
        materialized = node.get('config', {}).get('materialized')
        if materialized == 'ephemeral':
            return False
        return True

    is_chosen_fn = profiling_chosen_fn if not select_for_metadata else metadata_chosen_fn

    for key, node in nodes.items():
        if node.get('resource_type') not in resource_whitelist:
            continue
        Statistics().add_field_one('total')
        if not is_chosen_fn(key, node):
            continue
        candidate.append(node)

    for key, node in sources.items():
        if not is_chosen_fn(key, node):
            continue
        candidate.append(node)

    return candidate


def get_dbt_state_tests_result(dbt_state_dir: str, table_filter=None):
    output = []
    unique_tests = {}

    run_results = _get_state_run_results(dbt_state_dir)
    manifest = _get_state_manifest(dbt_state_dir)

    nodes = manifest.get('nodes')
    sources = manifest.get('sources')
    for result in run_results.get('results', []):
        unique_id = result.get('unique_id')

        node = nodes.get(unique_id)
        if not node:
            continue

        if node.get('resource_type') != 'test':
            continue

        # The test is just compiled, but not executed
        if result.get('status') == 'success':
            continue

        unique_tests[unique_id] = dict(
            status=result.get('status'),
            failures=result.get('failures'),
            message=result.get('message'),
        )

        test_node = node
        table = None
        depends_on_nodes = test_node.get('depends_on', {}).get('nodes', [])
        for depends_on_node_id in depends_on_nodes:
            depends_on_node = nodes.get(depends_on_node_id)
            if depends_on_node_id.startswith('source'):
                source = sources.get(depends_on_node_id)
                table = f"{source.get('source_name')}.{source.get('name')}"
                break
            elif depends_on_node.get('resource_type') in ['model', 'seed']:
                table = depends_on_node.get('name')
                break
        column = test_node.get('column_name')

        if table is None:
            continue

        if table_filter is not None:
            if len(table.split('.')) == 2:
                _, table_name = table.split('.')
            else:
                table_name = table
            if table_name != table_filter:
                continue

        output.append(dict(
            id=unique_id,
            name=unique_id,
            table=table,
            column=column if column != test_node['name'] else None,
            status='failed' if unique_tests[unique_id]['status'] == 'fail' else 'passed',
            tags=[],
            message=unique_tests[unique_id]['message'],
            display_name=test_node['name'],
            source='dbt'
        ))

    return output


def get_dbt_state_metrics(dbt_state_dir: str, dbt_tag: str, dbt_resources: Optional[dict] = None):
    manifest = _get_state_manifest(dbt_state_dir)

    def is_chosen(key, metric):
        statistics = Statistics()
        if dbt_resources:
            chosen = key in dbt_resources['metrics']
            if not chosen:
                statistics.add_field_one('filter')
            return chosen
        if dbt_tag is not None:
            chosen = dbt_tag in metric.get('tags')
            if not chosen:
                statistics.add_field_one('notag')
            return chosen
        return True

    metrics = []
    metric_map = {}
    for key, metric in manifest.get('metrics').items():
        if metric.get('calculation_method') == 'derived':
            table = None
            schema = None
            database = None
        else:
            nodes = metric.get('depends_on').get('nodes', [])
            # Note: In "metrics" definition, if "model" value doesn't use `ref()` function,
            #       the "depends_on.nodes" will be empty in the generated metrics manifest
            if len(nodes) == 0:
                continue
            depends_on_node = nodes[0]
            table = manifest.get('nodes').get(depends_on_node).get('alias')
            schema = manifest.get('nodes').get(depends_on_node).get('schema')
            database = manifest.get('nodes').get(depends_on_node).get('database')

        m = Metric(metric.get('name'), table, schema, database, metric.get('expression'), metric.get('timestamp'),
                   metric.get('calculation_method'), metric.get('time_grains'), dimensions=None,
                   filters=metric.get('filters'), label=metric.get('label'), description=metric.get('description'),
                   ref_id=metric.get('unique_id'))

        metric_map[key] = m
        statistics = Statistics()
        statistics.add_field_one('total')

        if is_chosen(key, metric):
            if metric.get('window'):
                console.print(
                    f"[[bold yellow]Warning[/bold yellow]] Skip metric '{metric.get('name')}'. Property 'window' is not supported.")
                statistics.add_field_one('nosupport')
                continue
            metrics.append(m)

    for key, metric in metric_map.items():
        if metric.calculation_method == 'derived':
            for depends_on_metric in manifest.get('metrics').get(key).get('depends_on').get('nodes'):
                metric.ref_metrics.append(metric_map.get(depends_on_metric))

    return metrics


def is_dbt_schema_version_16(manifest: Dict):
    # dbt_schema_version: 'https://schemas.getdbt.com/dbt/manifest/v10.json'
    schema_version = manifest['metadata'].get('dbt_schema_version').split('/')[-1]
    version = schema_version.split('.')[0]
    return int(version[1:]) >= 10


def load_metric_jinja_string_template(value: str):
    from jinja2 import Environment, BaseLoader
    env = Environment(loader=BaseLoader())

    def dimension(var):
        return var

    env.globals['Dimension'] = dimension
    template = env.from_string(value)

    return template


def get_support_time_grains(grain: str):
    all_time_grains = ['day', 'week', 'month', 'quarter', 'year']
    available_time_grains = all_time_grains[all_time_grains.index(grain):]
    support_time_grains = ['day', 'month', 'year']

    return [x for x in support_time_grains if x in available_time_grains]


def find_derived_time_grains(manifest: Dict, metric: Dict):
    nodes = metric.get('depends_on').get('nodes', [])
    depends_on = nodes[0]
    if depends_on.startswith('semantic_model.'):
        semantic_model = manifest.get('semantic_models').get(depends_on)
        measure = None
        for obj in semantic_model.get('measures'):
            if obj.get('name') == metric.get('type_params').get('measure').get('name'):
                measure = obj
                break

        agg_time_dimension = measure.get('agg_time_dimension') if measure.get(
            'agg_time_dimension') else semantic_model.get('defaults').get('agg_time_dimension')

        time_grains = None
        if agg_time_dimension is not None:
            # find dimension definition - time
            for obj in semantic_model.get('dimensions'):
                if obj.get('name') == agg_time_dimension:
                    grain = obj.get('type_params').get('time_granularity')
                    time_grains = get_support_time_grains(grain)
                    break

        return time_grains


def get_dbt_state_metrics_16(dbt_state_dir: str, dbt_tag: Optional[str] = None, dbt_resources: Optional[dict] = None):
    manifest = _get_state_manifest(dbt_state_dir)

    if not is_dbt_schema_version_16(manifest):
        console.print("[[bold yellow]Skip[/bold yellow]] Metric query is not supported for dbt < 1.6")
        return []

    def is_chosen(key, metric):
        statistics = Statistics()
        if dbt_resources:
            chosen = key in dbt_resources['metrics']
            if not chosen:
                statistics.add_field_one('filter')
            return chosen
        if dbt_tag is not None:
            chosen = dbt_tag in metric.get('tags')
            if not chosen:
                statistics.add_field_one('notag')
            return chosen
        return True

    metrics = []
    metric_map = {}

    for key, metric in manifest.get('metrics').items():
        metric_map[metric.get('name')] = metric

    def _create_metric(name, filter=None, alias=None):
        statistics = Statistics()
        metric = metric_map.get(name)

        if metric.get('type') == 'simple':
            primary_entity = None
            metric_filter = []
            if metric.get('filter') is not None:
                sql_filter = load_metric_jinja_string_template(metric.get('filter').get('where_sql_template')).render()
                metric_filter.append({'field': sql_filter.split(' ')[0],
                                      'operator': sql_filter.split(' ')[1],
                                      'value': sql_filter.split(' ')[2]})
            if filter is not None:
                sql_filter = load_metric_jinja_string_template(filter.get('where_sql_template')).render()
                metric_filter.append({'field': sql_filter.split(' ')[0],
                                      'operator': sql_filter.split(' ')[1],
                                      'value': sql_filter.split(' ')[2]})

            nodes = metric.get('depends_on').get('nodes', [])
            depends_on = nodes[0]
            if depends_on.startswith('semantic_model.'):
                semantic_model = manifest.get('semantic_models').get(depends_on)
                table = semantic_model.get('node_relation').get('alias')
                schema = semantic_model.get('node_relation').get('schema_name')
                database = semantic_model.get('node_relation').get('database')
                # find measure definition
                measure = None
                for obj in semantic_model.get('measures'):
                    if obj.get('name') == metric.get('type_params').get('measure').get('name'):
                        measure = obj
                        break
                # TODO: remove assertion
                assert measure is not None, 'Measure not found'
                expression = measure.get('expr') if measure.get('expr') is not None else measure.get('name')
                calculation_method = measure.get('agg')

                for entity in semantic_model.get('entities'):
                    if entity.get('type') == 'primary':
                        primary_entity = entity.get('name')
                        break

                agg_time_dimension = measure.get('agg_time_dimension') if measure.get(
                    'agg_time_dimension') else semantic_model.get('defaults').get('agg_time_dimension')
                timestamp = None
                time_grains = None
                if agg_time_dimension is not None:
                    # find dimension definition - time
                    for obj in semantic_model.get('dimensions'):
                        if obj.get('name') == agg_time_dimension:
                            timestamp = obj.get('expr')
                            grain = obj.get('type_params').get('time_granularity')
                            time_grains = get_support_time_grains(grain)
                            break

                if metric.get('filter') is not None:
                    # find dimension definition - categorical
                    for obj in semantic_model.get('dimensions'):
                        if obj.get('name') in metric_filter[0]['field']:
                            expr = obj.get('expr') or obj.get('name')
                            metric_filter[0]['field'] = metric_filter[0]['field'].replace(obj.get('name'), expr)
                            break
            else:
                # TODO: remove assertion
                assert False, 'Simple type metric should depend on semantic model.'

            model = SemanticModel(metric.get('name'), table, schema, database, expression, timestamp,
                                  filters=metric_filter)

            m = Metric(metric.get('name'), model=model, calculation_method=calculation_method, time_grains=time_grains,
                       label=metric.get('label'), description=metric.get('description'),
                       ref_id=metric.get('unique_id'))

            for f in m.model.filters:
                if primary_entity in f['field']:
                    f['field'] = f['field'].replace(f'{primary_entity}__', '')
                else:
                    console.print(
                        f"[[bold yellow]Skip[/bold yellow]] Metric '{metric.get('name')}'. "
                        f"Dimension of foreign entities is not supported.")
                    statistics.add_field_one('nosupport')
                    return None
            if m.calculation_method == 'median':
                console.print(
                    f"[[bold yellow]Skip[/bold yellow]] Metric '{metric.get('name')}'. "
                    f"Aggregation type 'median' is not supported.")
                statistics.add_field_one('nosupport')
                return None

            return m
        elif metric.get('type') == 'derived':
            ref_metrics = []
            time_grains = ['day', 'month', 'year']
            expr = metric.get('type_params').get('expr')
            for ref_metric in metric.get('type_params').get('metrics'):
                if ref_metric.get('offset_window') is not None:
                    console.print(
                        f"[[bold yellow]Skip[/bold yellow]] Metric '{metric.get('name')}'. "
                        f"Derived metric property 'offset_window' is not supported.")
                    statistics.add_field_one('nosupport')
                    return None
                m2 = _create_metric(
                    ref_metric.get('name'),
                    filter=ref_metric.get('filter'),
                    alias=ref_metric.get('alias'))
                ref_metrics.append(m2)

                derived_time_grains = find_derived_time_grains(manifest, metric_map[ref_metric.get('name')])
                if len(time_grains) < len(derived_time_grains):
                    time_grains = derived_time_grains

                if ref_metric.get('alias') is not None:
                    expr = expr.replace(ref_metric.get('alias'), ref_metric.get('name'))

            m = Metric(metric.get('name'),
                       calculation_method='derived',
                       expression=expr,
                       time_grains=time_grains,
                       label=metric.get('label'), description=metric.get('description'), ref_metrics=ref_metrics,
                       ref_id=metric.get('unique_id'))
            return m
        elif metric.get('type') == 'ratio':
            ref_metrics = []
            time_grains = ['day', 'month', 'year']
            numerator = metric.get('type_params').get('numerator')
            m2 = _create_metric(
                numerator.get('name'),
                filter=numerator.get('filter'),
                alias=numerator.get('alias'))
            ref_metrics.append(m2)
            derived_time_grains = find_derived_time_grains(manifest, metric_map[numerator.get('name')])
            if len(time_grains) < len(derived_time_grains):
                time_grains = derived_time_grains

            denominator = metric.get('type_params').get('denominator')
            m2 = _create_metric(
                denominator.get('name'),
                filter=denominator.get('filter'),
                alias=denominator.get('alias'))
            ref_metrics.append(m2)
            derived_time_grains = find_derived_time_grains(manifest, metric_map[denominator.get('name')])
            if len(time_grains) < len(derived_time_grains):
                time_grains = derived_time_grains

            m = Metric(metric.get('name'),
                       calculation_method='derived',
                       expression=f"{numerator.get('name')} / {denominator.get('name')}",
                       time_grains=time_grains,
                       label=metric.get('label'), description=metric.get('description'), ref_metrics=ref_metrics,
                       ref_id=metric.get('unique_id'))
            return m
        else:
            console.print(
                f"[[bold yellow]Skip[/bold yellow]] Metric '{metric.get('name')}'. "
                f"Metric type 'Cumulative' is not supported.")
            statistics.add_field_one('nosupport')
        return None

    for key, metric in manifest.get('metrics').items():
        statistics = Statistics()
        statistics.add_field_one('total')

        if not is_chosen(key, metric):
            continue

        m = _create_metric(metric.get('name'))
        if m is not None:
            metrics.append(m)

    return metrics


def check_dbt_manifest(dbt_state_dir: str) -> bool:
    path = os.path.join(dbt_state_dir, 'manifest.json')
    if os.path.isabs(path) is False:
        from piperider_cli.configuration import FileSystem
        path = os.path.join(FileSystem.WORKING_DIRECTORY, path)
    return os.path.exists(path)


def get_dbt_manifest(dbt_state_dir: str):
    return _get_state_manifest(dbt_state_dir)


def load_dbt_resources(target_path: str, select: tuple = None, state=None):
    manifest = load_manifest(get_dbt_manifest(target_path)) if state is None else load_full_manifest(target_path)
    try:
        list_resources = list_resources_unique_id_from_manifest(manifest, select=select, state=state)
    except RuntimeError as e:
        raise DbtRunTimeError(e, select, state)
    return read_dbt_resources(list_resources)


def get_dbt_run_results(dbt_state_dir: str):
    return _get_state_run_results(dbt_state_dir) if is_dbt_run_results_ready(dbt_state_dir) else None


def load_dbt_project(path: str):
    """
    Load dbt project file and return the content of 'profile' and 'target-path' fields
    """
    if not path.endswith('dbt_project.yml'):
        path = os.path.join(path, 'dbt_project.yml')

    if not os.path.isabs(path):
        from piperider_cli.configuration import FileSystem
        if os.path.exists(path) is False:
            path = os.path.join(FileSystem.WORKING_DIRECTORY, path)

    with open(path, 'r') as fd:
        try:
            yml = yaml.YAML()
            yml.allow_duplicate_keys = True
            dbt_project = yml.load(fd)

            content = {}
            for key, val in dbt_project.items():
                if key not in ['profile', 'target-path']:
                    continue
                if isinstance(val, str):
                    content[key] = load_jinja_string_template(val).render()
                else:
                    content[key] = val
            return content
        except Exception as e:
            raise DbtProjectInvalidError(path, e)


def load_dbt_profile(path):
    template = load_jinja_template(path)
    profile = None
    try:
        yml = yaml.YAML()
        yml.allow_duplicate_keys = True
        profile = yml.load(template.render())
    except Exception as e:
        raise DbtProfileInvalidError(path, e)
    if profile is None:
        raise DbtProfileInvalidError(path, f"The profile '{path}' is empty")
    return profile


def load_credential_from_dbt_profile(dbt_profile, profile_name, target_name):
    credential = dbt_profile.get(profile_name, {}).get('outputs', {}).get(target_name, {})

    if credential.get('type') == 'bigquery':
        # BigQuery Data Source
        from piperider_cli.datasource.bigquery import AUTH_METHOD_OAUTH_SECRETS
        # DBT profile support 4 types of methods to authenticate with BigQuery:
        #   [ 'oauth', 'oauth-secrets', 'service-account', 'service-account-json' ]
        # Ref: https://docs.getdbt.com/reference/warehouse-profiles/bigquery-profile#authentication-methods
        if credential.get('method') == 'oauth-secrets':
            credential['method'] = AUTH_METHOD_OAUTH_SECRETS
            # TODO: Currently SqlAlchemy haven't support using access token to authenticate with BigQuery.
            #       Ref: https://github.com/googleapis/python-bigquery-sqlalchemy/pull/459
            raise DbtProfileBigQueryAuthWithTokenUnsupportedError
    elif credential.get('type') == 'redshift':
        if credential.get('method') is None:
            credential['method'] = 'password'
        host = credential.get('host')
        port = credential.get('port')
        dbname = credential.get('dbname')
        credential['endpoint'] = f'{host}:{port}/{dbname}'
    return credential


def read_dbt_resources(source: Union[str, io.TextIOWrapper, list]):
    if isinstance(source, io.TextIOWrapper):
        lines = source.readlines()
    elif isinstance(source, str):
        lines = source.split('\n')
    else:
        lines = source

    metrics = []
    models = []
    for line in lines:
        dbt_resource = line.rstrip()
        if ' ' in dbt_resource:
            # From dbt 1.5.x, `dbt list` will output runtime logs as well. Need to ignore them
            continue
        elif dbt_resource.startswith('source.'):
            continue
        elif dbt_resource.startswith('metric.'):
            metrics.append(dbt_resource)
        else:
            models.append(dbt_resource)
    return dict(metrics=metrics, models=models)


def prepare_topological_graph(manifest: Dict):
    child_map = manifest.get('child_map', {})
    graph = {}
    for k, v in child_map.items():
        if k.split('.')[0] == 'model' or k.split('.')[0] == 'metric':
            v = [x for x in v if x.split('.')[0] == 'model' or x.split('.')[0] == 'metric']
            graph[k] = v

    return graph



from io import StringIO
from unittest import TestCase

from ruamel import yaml

from piperider_cli.assertion_engine import AssertionEngine


def _(assertion_content: str):
    def content_provider(self):
        self.assertions_content = yaml.safe_load(StringIO(assertion_content))
        return [], []

    return content_provider


class BuiltinValidationColumnTypesTests(TestCase):

    def setUp(self) -> None:
        self.engine = AssertionEngine(None)
        self.origin_function = AssertionEngine.load_assertion_content

    def tearDown(self) -> None:
        AssertionEngine.load_assertion_content = self.origin_function

    def test_validation_assert_column_schema_type(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_schema_type
                    assert:
                      schema_type: JSON
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

        # test with invalid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_schema_type
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(results[0].as_internal_report(), r"""
Found assertion syntax problem => name: assert_column_schema_type for table column_types and column column_a
ERROR: schema_type parameter is required
ERROR: schema_type parameter should be a <class 'str'> value
        """.strip())

    def test_validation_assert_column_type(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_type
                    assert:
                      type: string
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

        # test with invalid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_type
                    assert:
                      type: foobarbar
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(results[0].as_internal_report(), r"""
Found assertion syntax problem => name: assert_column_type for table column_types and column column_a
ERROR: type parameter should be one of ['string', 'integer', 'numeric', 'datetime', 'date', 'time', 'boolean', 'other'], input: foobarbar
        """.strip())

    def test_validation_assert_column_in_types(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_in_types
                    assert:
                      types: [string]
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

        # test with invalid format
        AssertionEngine.load_assertion_content = _("""
            column_types:
              columns:
                column_a:
                  tests:
                  - name: assert_column_in_types
                    assert:
                      types: [foobarbar]
                    tags:
                    - OPTIONAL
            """)

        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(results[0].as_internal_report(), r"""
Found assertion syntax problem => name: assert_column_in_types for table column_types and column column_a
ERROR: types parameter should be one of ['string', 'integer', 'numeric', 'datetime', 'date', 'time', 'boolean', 'other'], input: ['foobarbar']
        """.strip())



from unittest import TestCase

from piperider_cli.compare_report import join, ComparisonData, _merge_keys


class CompareReportTests(TestCase):

    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_join(self):
        base = {
            'table_a': {
                'row_count': 50,
                'columns': {
                    'col_a': {
                        'samples': 5
                    }
                }
            },
            'table_b': {
                'row_count': 100,
                'columns': {
                    'col_a': {
                        'samples': 10
                    }
                }
            }
        }

        target = {
            'table_a': {
                'row_count': 50,
                'columns': {
                    'col_a': {
                        'samples': 10
                    }
                }
            },
            'table_b': {
                'row_count': 100,
                'columns': {
                    'col_a': {
                        'samples': 5
                    }
                }
            },
            'table_c': {
                'row_count': 150,
                'columns': {
                    'col_a': {
                        'samples': 15
                    }
                }
            }
        }

        result = join(base, target)
        self.assertEqual(len(result.keys()), 3)
        self.assertListEqual(list(result.keys()), ['table_a', 'table_b', 'table_c'])
        self.assertEqual(len(result['table_a'].keys()), 2)

        joined_table = result['table_b']
        columns_b = joined_table.get('base', {}).get('columns') if joined_table.get('base') else None
        columns_t = joined_table.get('target', {}).get('columns') if joined_table.get('target') else None
        joined_columns = join(columns_b, columns_t)
        joined_column = joined_columns['col_a']

        b = joined_column.get('base')
        t = joined_column.get('target')

        samples_b = b.get('samples', 0) if b else None
        samples_t = t.get('samples', 0) if t else None

        self.assertEqual(samples_b, 10)
        self.assertEqual(samples_t, 5)

        joined_table = result['table_c']
        columns_b = joined_table.get('base', {}).get('columns') if joined_table.get('base') else None
        columns_t = joined_table.get('target', {}).get('columns') if joined_table.get('target') else None
        joined_columns = join(columns_b, columns_t)
        joined_column = joined_columns['col_a']

        b = joined_column.get('base')
        t = joined_column.get('target')

        samples_b = b.get('samples', 0) if b else None
        samples_t = t.get('samples', 0) if t else None

        self.assertEqual(samples_b, None)
        self.assertEqual(samples_t, 15)

    def test_merge_keys(self):
        self.assertEqual(['a', 'b', 'c', 'd', 'e'], _merge_keys(['a', 'b', 'c', 'd'], ['a', 'b', 'c', 'e']))
        self.assertEqual(['a', 'b', 'c', 'd', 'e'], _merge_keys(['a', 'b', 'c', 'e'], ['a', 'b', 'd', 'e']))
        self.assertEqual(['a', 'z', 'b', 'c', 'd'], _merge_keys(['a', 'b', 'c', 'd'], ['z', 'b', 'c', 'd']))
        self.assertEqual(['a', 'b', 'c', 'd', 'e'], _merge_keys(['a', 'b', 'c', 'd'], ['a', 'd', 'e']))
        self.assertEqual(['a', 'b', 'z', 'c', 'd'], _merge_keys(['a', 'b', 'c', 'd'], ['a', 'b', 'z', 'c', 'd']))
        self.assertEqual(['z', 'a', 'b', 'c', 'd'], _merge_keys(['a', 'b', 'c', 'd'], ['z', 'a', 'b', 'c', 'd']))
        self.assertEqual(['a', 'b', 'c', 'd'], _merge_keys(['a', 'b'], ['c', 'd']))



import hashlib
import os.path


def schema_version():
    import piperider_cli.profiler as p
    schema_def = os.path.join(os.path.dirname(p.__file__), 'schema.json')
    with open(schema_def) as fh:
        m = hashlib.md5()
        m.update(fh.read().encode())
        return m.hexdigest()


if __name__ == '__main__':
    print(schema_version())



class ProfilerEventHandler:
    def handle_run_start(self, run_result):
        pass

    def handle_run_end(self, run_result):
        pass

    def handle_run_progress(self, run_result, total, completed):
        pass

    def handle_metadata_start(self):
        pass

    def handle_metadata_progress(self, total, completed):
        pass

    def handle_metadata_end(self):
        pass

    def handle_manifest_start(self):
        pass

    def handle_manifest_progress(self, total, completed):
        pass

    def handle_manifest_end(self):
        pass

    def handle_table_start(self, table_name):
        pass

    def handle_table_progress(self, table_name, table_result, total, completed):
        pass

    def handle_table_end(self, table_name, table_result):
        pass

    def handle_column_start(self, table_name, column_name):
        pass

    def handle_column_end(self, table_name, column_name, column_result):
        pass


class DefaultProfilerEventHandler(ProfilerEventHandler):
    table_completed = 0
    table_total = 0
    col_completed = 0
    col_total = 0

    def handle_run_start(self, run_result):
        print("Start profiling")

    def handle_run_progress(self, run_result, total, completed):
        self.table_total = total
        self.table_completed = completed

    def handle_run_end(self, run_result):
        pass

    def handle_metadata_start(self):
        pass

    def handle_metadata_progress(self, total, completed):
        pass

    def handle_metadata_end(self):
        pass

    def handle_manifest_start(self):
        pass

    def handle_manifest_progress(self, total, completed):
        pass

    def handle_manifest_end(self):
        pass

    def handle_table_start(self, table_name):
        pass

    def handle_table_progress(self, table_name, table_result, total, completed):
        self.col_total = total
        self.col_completed = completed
        if completed == 0:
            print(
                f"[{self.table_completed + 1}/{self.table_total}] profiling [{table_name}] rows={table_result['row_count']}")

    def handle_table_end(self, table_name, table_result):
        pass

    def handle_column_start(self, table_name, column_name):
        pass

    def handle_column_end(self, table_name, column_name, column_result):
        print(
            f"    [{self.col_completed + 1}/{self.col_total}] profiling [{table_name}.{column_name}] type={column_result['schema_type']} [{column_result['elapsed_milli']}ms]")



from typing import Any, List


class MarkdownTable:

    def __init__(self, *, headers: List[str]):
        self.headers: List[str] = headers
        self.num_of_columns = len(self.headers)
        self.rows = []

    def add_row(self, columns: List):
        if len(columns) != self.num_of_columns:
            raise ValueError('The number of columns doesn\'t match headers')

        self.rows.append(columns)

    def build(self):
        header_lines = "| " + " | ".join(self.headers).strip()
        header_ending = "| " + " | ".join(["--" for x in self.headers]).strip()
        row_lines = []

        for r in self.rows:
            row = "| " + " | ".join([self.as_str(x) for x in r]).strip()
            row_lines.append(row)

        return "\n".join([header_lines, header_ending] + row_lines)

    def as_str(self, entry: Any):
        if isinstance(entry, str):
            return entry
        return str(entry)



import json
import os

import requests


def fetch_pr_metadata() -> dict:
    '''
    If piperider is running in a GitHub Action, this function will return the pull request metadata.

    Example:
    {
        "github_pr_id": 1,
        "github_pr_url": "https://github.com/xyz/abc/pull/1
        "github_pr_title": "Update README.md"
    }

    :return: dict
    '''

    # get the event json from the path in GITHUB_EVENT_PATH
    event_path = os.getenv("GITHUB_EVENT_PATH")
    if event_path:
        try:
            with open(event_path, "r") as event_file:
                event_data = json.load(event_file)

            pr_id = event_data["number"]
            if event_data.get("pull_request"):
                pull_request_data = event_data["pull_request"]
                pr_url = pull_request_data["_links"]["html"]["href"]
                pr_api = pull_request_data["_links"]["self"]["href"]
                pr_title = _fetch_pr_title(pr_api)
                return dict(github_pr_id=pr_id, github_pr_url=pr_url, github_pr_title=pr_title)
            else:
                print("Not a pull request event, skip.")
        except Exception as e:
            print("Cannot parse github action event", e)
    return None


def _fetch_pr_title(endpoint) -> str:
    github_token = os.getenv("GITHUB_TOKEN")

    if github_token is None:
        return None

    try:
        headers = {"Authorization": f"Bearer {github_token}"}
        response = requests.get(endpoint, headers=headers)

        if response.status_code == 200:
            pull_request_data = response.json()
            return pull_request_data.get('title')
    except Exception as e:
        print("Cannot fetch PR title: ", e)

    return None



from piperider_cli.datasource import DataSource
from piperider_cli.error import PipeRiderConnectorUnsupportedError


class UnsupportedDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'unsupported', **kwargs)
        self.fields = [
        ]

    def validate(self):
        return True, []

    def to_database_url(self):
        return ''

    def verify_connector(self):
        datasource_type = self.credential.get('type', 'Unknown')
        return PipeRiderConnectorUnsupportedError(
            f"Piperider doesn't support data source '{self.name}' of type '{datasource_type}'.",
            datasource_type)

    def _get_display_description(self):
        return ''

    def get_database(self):
        return ''

    def get_schema(self):
        return ''

    def get_engine_by_database(self):
        return None



import json
import os
from typing import List

import inquirer

from piperider_cli.error import PipeRiderConnectorError
from . import DataSource
from .field import PathField, ListField, DataSourceField, _default_validate_func

APPLICATION_DEFAULT_CREDENTIALS = os.path.join(os.path.expanduser('~'), '.config', 'gcloud',
                                               'application_default_credentials.json')
DEFAULT_GCP_CREDENTIALS = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', APPLICATION_DEFAULT_CREDENTIALS)
SUGGESTION_FOR_OAUTH_FAILED = f'Default Credential file {DEFAULT_GCP_CREDENTIALS} was not found.\n' \
                              'Please execute the following command to generate it.\n' \
                              '[green]' \
                              'gcloud auth application-default login \\\n' \
                              '--scopes=https://www.googleapis.com/auth/bigquery,\\\n' \
                              'https://www.googleapis.com/auth/drive.readonly,\\\n' \
                              'https://www.googleapis.com/auth/iam.test' \
                              '[/green]'
AUTH_METHOD_OAUTH = 'oauth'
AUTH_METHOD_OAUTH_SECRETS = 'oauth_secrets'
AUTH_METHOD_SERVICE_ACCOUNT = 'service-account'
AUTH_METHOD_SERVICE_ACCOUNT_JSON = 'service-account-json'


class HiddenProjectListFromOAuthField(DataSourceField):

    def __init__(self, name, ds: DataSource):
        super().__init__(name, "hidden", description=f'{name}')
        self.ds = ds
        self.callback = self.list_oauth_projects

    def question(self):
        question = inquirer.Text(self.name)
        question.callback = self.list_oauth_projects
        return question

    def list_oauth_projects(self, answers: dict):
        try:
            import google.auth
            import google.oauth2.service_account
            from google.cloud.bigquery.client import Client as BqClient
            from google.cloud.bigquery.dataset import DatasetListItem

            credentials, project_id = None, None
            if answers['method'] == AUTH_METHOD_OAUTH:
                credentials, project_id = google.auth.load_credentials_from_file(DEFAULT_GCP_CREDENTIALS)
            elif answers.get('keyfile', None) is not None:
                credentials, project_id = google.auth.load_credentials_from_file(os.path.expanduser(answers['keyfile']))

            if project_id is None and hasattr(credentials, 'quota_project_id'):
                project_id = getattr(credentials, 'quota_project_id')

            def list_dataset(project_id):
                client = BqClient(project=project_id, credentials=credentials)
                for x in client.list_datasets(project_id):
                    item: DatasetListItem = x
                    yield json.dumps(dict(project=item.project, dataset=item.dataset_id))

            def list_projects(credentials):
                import google.cloud.client as common_client
                from google.api_core.client_options import ClientOptions
                options = ClientOptions(scopes=['https://www.googleapis.com/auth/cloudplatformprojects.readonly'])
                client = common_client.Client(credentials=credentials, client_options=options)
                list_project = 'https://cloudresourcemanager.googleapis.com/v1/projects'
                response = client._http.get(list_project).json()
                if 'error' in response:
                    raise ValueError(response)
                for p in response['projects']:
                    yield p['projectId']

            # if project_id was found, it could be a service account
            if project_id:
                outputs = list(list_dataset(project_id))
                self.ds.available_datasets += outputs
                return None
            else:
                outputs = []
                for p in list_projects(credentials):
                    outputs += list(list_dataset(p))
                    self.ds.available_datasets += outputs
                return None

        except BaseException as e:
            if answers['method'] == AUTH_METHOD_OAUTH:
                from rich.console import Console
                console = Console()
                console.print(SUGGESTION_FOR_OAUTH_FAILED)
            raise e


class HiddenReduceField(DataSourceField):

    def __init__(self, name, ds: DataSource):
        super().__init__(name, "hidden", description=f'{name}')
        self.ds = ds
        self.callback = self.reformat_answers

    def question(self):
        question = inquirer.Text(self.name)
        question.callback = self.reformat_answers
        return question

    def reformat_answers(self, answers: dict):
        selection = json.loads(answers['selected_dataset'])
        return selection[self.name]


def validate_service_file(answer, current) -> bool:
    if _default_validate_func(answer, current):
        valid_service_account = check_json_file_has_service_account_type(current)
        if valid_service_account:
            return True
        else:
            from rich.console import Console
            console = Console()
            console.print(f'    [[red]Error[/red]] Not a valid service-account json file')  # noqa: F541
            return False
    return False


def check_json_file_has_service_account_type(filepath):
    try:
        with open(os.path.expanduser(filepath)) as fh:
            service_account_info = json.loads(fh.read())
            if service_account_info.get('type') == 'service_account':
                return True
            else:
                return False
    except BaseException:
        return False


class SystemAccountPathField(DataSourceField):
    def __init__(self, name, value=None, default=None, description=None, validate=_default_validate_func,
                 optional=False, ignore=False):
        super().__init__(name, "path", value, default, description, validate, optional, ignore)

    def question(self):
        path = inquirer.Path(self.name, message=self.description, default=self.default, exists=True, ignore=self.ignore)
        path.origin_validate = path.validate

        def wrapper_validate(current):
            # check the original validator
            path.origin_validate(current)
            # verify the file is a service account file
            filepath = path.normalize_value(current)
            if not check_json_file_has_service_account_type(filepath):
                from inquirer import errors
                raise errors.ValidationError(current, reason='Not a valid service-account json file')

        path.validate = wrapper_validate
        return path


class BigQueryDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'bigquery', **kwargs)

        error = self.verify_connector()
        if error:
            raise error

        self.available_datasets: List[str] = []
        self.fields = [
            ListField('method', description='Authentication Methods',
                      default=[AUTH_METHOD_OAUTH, AUTH_METHOD_SERVICE_ACCOUNT]),
            SystemAccountPathField('keyfile', description='The path of GCP Service Account Key File',
                                   ignore=lambda answers: AUTH_METHOD_SERVICE_ACCOUNT != answers['method'],
                                   validate=validate_service_file),
            HiddenProjectListFromOAuthField('selected_dataset', self),
            ListField('selected_dataset', description='Pick a dataset',
                      default=self.available_datasets),
            HiddenReduceField('dataset', self),
            HiddenReduceField('project', self),
        ]

    def validate(self):
        reasons = []
        if self.type_name != 'bigquery':
            raise ValueError('type name should be bigquery')

        method = self.credential.get('method')
        if method not in [AUTH_METHOD_OAUTH, AUTH_METHOD_SERVICE_ACCOUNT,
                          AUTH_METHOD_SERVICE_ACCOUNT_JSON]:
            reasons.append(f'Unsupported authentication method: {method}')

        if method == AUTH_METHOD_OAUTH:
            if not os.path.exists(DEFAULT_GCP_CREDENTIALS):
                reasons.append(SUGGESTION_FOR_OAUTH_FAILED)
        if method == AUTH_METHOD_SERVICE_ACCOUNT:
            if self.credential.get('keyfile') is None:
                reasons.append('keyfile is required when authentication method is service-account')
        if method == AUTH_METHOD_SERVICE_ACCOUNT_JSON:
            if self.credential.get('keyfile_json') is None:
                reasons.append('keyfile_json is required when authentication method is service-account-json')

        return reasons == [], reasons

    def to_database_url(self, database):
        from sqlalchemy_bigquery import BigQueryDialect

        BigQueryDialect.supports_statement_cache = True

        project = database
        if project is None:
            project = self.credential["project"]
        if project is None:
            project = self.credential["database"]

        dataset = self.credential.get("dataset")
        if dataset is None:
            dataset = self.credential.get("schema")

        return f'bigquery://{project}/{dataset}'

    def engine_args(self):
        args = dict()
        if self.credential.get('method') == AUTH_METHOD_SERVICE_ACCOUNT:
            args['credentials_path'] = self.credential.get('keyfile')
        elif self.credential.get('method') == AUTH_METHOD_SERVICE_ACCOUNT_JSON:
            args['credentials_info'] = self.credential.get('keyfile_json', {})
        return args

    def verify_connector(self):
        try:
            from sqlalchemy_bigquery import BigQueryDialect

            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'bigquery')

    def _get_display_description(self):
        cred = self.credential
        project = cred.get('project') if cred.get('project') else cred.get('database')
        dataset = cred.get('dataset') if cred.get('dataset') else cred.get('schema')
        return f"type={self.type_name}, project={project}, dataset={dataset}"

    def get_database(self):
        cred = self.credential
        project = cred.get('project') if cred.get('project') else cred.get('database')
        return project

    def get_schema(self):
        cred = self.credential
        dataset = cred.get('dataset') if cred.get('dataset') else cred.get('schema')
        return dataset



import os
from unittest import TestCase, mock

from piperider_cli import load_jinja_string_template


class TestUtil(TestCase):

    @mock.patch.dict(os.environ, {"dummy_env_text": "test", "dummy_env_number": str(1), "dummy_env_bool": str(True)})
    def test_load_jinja_string_template(self):
        val = "{{ env_var('dummy_env_text') | as_text }}"
        result = load_jinja_string_template(val).render()

        self.assertEqual("test", result)

        val = "{{ env_var('dummy_env_number') | as_number }}"
        result = load_jinja_string_template(val).render()

        self.assertEqual('1', result)

        val = "{{ env_var('dummy_env_bool') | as_bool }}"
        result = load_jinja_string_template(val).render()

        self.assertEqual("True", result)

        val = "{{ env_var('dummy_env_none') }}"
        result = load_jinja_string_template(val).render()

        self.assertEqual("", result)

        val = "{{ env_var('dummy_env_none') | as_number }}"
        result = load_jinja_string_template(val).render()

        self.assertEqual("", result)




import os
from unittest import TestCase, mock

from click.testing import CliRunner

from piperider_cli.cli import init, version, diagnose, run


class TestPipeRiderCli(TestCase):
    def setUp(self) -> None:
        self.cli_runner = CliRunner()
        self.unit_test_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mock_dbt_project')
        pass

    def test_version_command(self):
        result = self.cli_runner.invoke(version)
        assert result.exit_code == 0

    def test_init_command(self):
        result = self.cli_runner.invoke(init, ["--dbt-project-dir", self.unit_test_dir])
        assert result.exit_code == 0

    @mock.patch('piperider_cli.validator.Validator.diagnose', return_value=True)
    def test_diagnose_command(self, *args):
        result = self.cli_runner.invoke(diagnose, ["--dbt-project-dir", self.unit_test_dir])
        assert result.exit_code == 0

    @mock.patch('piperider_cli.runner.Runner.exec', return_value=0)
    @mock.patch('piperider_cli.generate_report.GenerateReport.exec', return_value=0)
    @mock.patch('piperider_cli.cloud_connector.CloudConnector.is_login', return_value=False)
    @mock.patch('piperider_cli.cloud_connector.CloudConnector.upload_latest_report', return_value=0)
    def test_run_command(self, *args):
        # Manually config dbt project and profiles dir
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--dbt-profiles-dir", self.unit_test_dir,
        ])
        assert result.exit_code == 0

        # Table and dbt-list cannot be used together
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--table", "test",
            "--dbt-list",
        ])
        assert result.exit_code == 1

        # Select and dbt-list cannot be used together
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--select", "test",
            "--dbt-list",
        ])
        assert result.exit_code == 1



import json
import os
import subprocess
import unittest
from unittest import TestCase

from packaging import version

from piperider_cli.compare_report import ComparisonData
from piperider_cli.dbt import dbt_version
from piperider_cli.error import PipeRiderError


def pbcopy_string(input_string):
    try:
        import platform
        if platform.system() != "Darwin":
            return

        subprocess.run(['pbcopy'], input=input_string.encode(), check=True)
        print("String copied to clipboard.")
    except Exception:
        pass


class TestCompareSummaryNG(TestCase):

    def setUp(self):
        self.maxDiff = None
        self.dbt_state_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data")

    def manifest_path(self, manifest_name):
        return os.path.join(self.dbt_state_dir, manifest_name)

    def manifest_dict(self, manifest_name):
        with open(self.manifest_path(manifest_name)) as fh:
            return json.loads(fh.read())

    @unittest.skipIf(dbt_version < version.parse('1.4'), 'this only works after manifests generated after the v1.4')
    def test_in_memory_compare_with_manifests(self):
        run1 = self.manifest_dict("jaffle_shop_base.json")
        run2 = self.manifest_dict("jaffle_shop_target.json")

        run1 = self.manifest_dict("sc-31587-with-ref-base.json")
        run2 = self.manifest_dict("sc-31587-with-ref-input.json")

        run1 = self.manifest_dict("sc-31723-base.json")
        run1 = self.manifest_dict("case_base_not_profiled_1.json")
        run2 = self.manifest_dict("sc-31723-target.json")

        data = ComparisonData(run1, run2, None)
        result = data.to_summary_markdown_ng()

        pbcopy_string(result)

        with open("output.md", "w") as fh:
            fh.write(result)

    def test_in_memory_compare_with_manifests_v_1_3(self):
        run1 = self.manifest_dict("jaffle_shop_base_1.3.json")
        run2 = self.manifest_dict("jaffle_shop_target_1.3.json")

        data = ComparisonData(run1, run2, None)
        result = data.to_summary_markdown_ng()

        with open("output_v1_3.md", "w") as fh:
            fh.write(result)

    @unittest.skipIf(dbt_version < version.parse('1.4'), 'this only works after manifests generated after the v1.4')
    def test_in_memory_compare_with_manifests_base_not_profiled(self):
        run1 = self.manifest_dict("case_base_not_profiled_1.json")
        run2 = self.manifest_dict("case_base_not_profiled_2.json")

        data = ComparisonData(run1, run2, None)
        result = data.to_summary_markdown_ng()

        with open("output.md", "w") as fh:
            fh.write(result)

    @unittest.skipIf(dbt_version < version.parse('1.4'), 'this only works after manifests generated after the v1.4')
    def test_in_memory_compare_with_manifests_target_not_profiled(self):
        run1 = self.manifest_dict("case_base_not_profiled_1.json")
        run2 = self.manifest_dict("case_base_not_profiled_2.json")

        data = ComparisonData(run2, run1, None)
        result = data.to_summary_markdown_ng()

        with open("output.md", "w") as fh:
            fh.write(result)

    @unittest.skipIf(dbt_version < version.parse('1.6'), 'before 1.6 the legacy metrics are supported')
    def test_in_memory_compare_with_manifests_v_1_6(self):
        run1 = self.manifest_dict("git-repo-analytics_metrics_1.json")
        run2 = self.manifest_dict("git-repo-analytics_metrics_2.json")

        with self.assertRaises(PipeRiderError) as ctx:
            data = ComparisonData(run1, run2, None)
        self.assertEqual("Found legacy metrics", ctx.exception.message)



import os
from pathlib import Path
from unittest import TestCase, mock, skip

import pytest

import piperider_cli.dbtutil as dbtutil
from piperider_cli.datasource.sqlite import SqliteDataSource
from piperider_cli.dbt import dbt_version
from piperider_cli.profiler import Profiler
from tests.common import create_table
from tests.test_dbt_manifest_compatible import _load_manifest


class TestRunner(TestCase):

    def setUp(self):
        self.dbt_state_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data")

    def test_get_dbt_state_candidate(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          tag=None))

        self.assertEqual(len(tables), 9)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[0].get('schema'), 'PUBLIC')
        self.assertEqual(tables[0].get('alias'), 'PRICE_PRESENT')
        self.assertEqual(tables[1].get('name'), 'event_filtered')
        self.assertEqual(tables[2].get('name'), 'session')
        self.assertEqual(tables[3].get('name'), 'event_enhanced')
        self.assertEqual(tables[4].get('name'), 'user')
        self.assertEqual(tables[5].get('name'), 'project_block')  # seed
        self.assertEqual(tables[6].get('name'), 'user_block')  # seed
        self.assertEqual(tables[7].get('name'), 'user_internal')  # seed
        self.assertEqual(tables[8].get('name'), 'project')

    def test_get_dbt_state_candidate_view_profile(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          tag=None))

        self.assertEqual(len(tables), 10)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[1].get('name'), 'event_filtered')
        self.assertEqual(tables[2].get('name'), 'stg_event')
        self.assertEqual(tables[3].get('name'), 'session')
        self.assertEqual(tables[4].get('name'), 'event_enhanced')
        self.assertEqual(tables[5].get('name'), 'user')
        self.assertEqual(tables[6].get('name'), 'project_block')  # seed
        self.assertEqual(tables[7].get('name'), 'user_block')  # seed
        self.assertEqual(tables[8].get('name'), 'user_internal')  # seed
        self.assertEqual(tables[9].get('name'), 'project')

    def test_get_dbt_state_candidate_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          tag=None))
        self.assertEqual(len(tables), 9)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')
        self.assertEqual(tables[0].get('schema'), 'PUBLIC')
        self.assertEqual(tables[0].get('alias'), 'PRICE_PRESENT')

    def test_get_dbt_state_candidate_resources_only_models(self):
        models = ['model.infusetude.session',
                  'model.infusetude.user']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'session')
        self.assertEqual(tables[1].get('name'), 'user')

    def test_get_dbt_state_candidate_resources_only_seeds(self):
        models = ['seed.infusetude.project_block',
                  'seed.infusetude.user_block']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 2)

    def test_get_dbt_state_candidate_resources_only_metrics(self):
        metrics = ['metric.infusetude.active_projects',
                   'metric.infusetude.active_projects_per_user',
                   'metric.infusetude.active_users',
                   'metric.infusetude.total_events',
                   'metric.infusetude.total_sessions',
                   'metric.infusetude.total_users']
        resources = dict(models=[], metrics=metrics)
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_resources_not_found_or_empty(self):
        models = ['infusetude.amplitude.notfound']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

        resources = dict(models=[], metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 0)

    def test_get_dbt_state_candidate_resources_without_view_profile(self):
        models = ['model.infusetude.stg_event',
                  'model.infusetude.session',
                  'model.infusetude.user']
        resources = dict(models=models, metrics=[])
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=resources,
                                                                          tag=None))
        self.assertEqual(len(tables), 3)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')
        self.assertEqual(tables[2].get('name'), 'user')

    def test_get_dbt_state_candidate_tag(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          tag='test'))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')

    def test_get_dbt_state_candidate_tag_and_view(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          tag='test'))
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].get('name'), 'stg_event')
        self.assertEqual(tables[1].get('name'), 'session')

    def test_get_dbt_state_candidate_tag_and_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          tag='piperider'))
        self.assertEqual(len(tables), 4)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')

    def test_get_dbt_state_candidate_tag_and_run_results_emptyset(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=None,
                                                                          dbt_resources=None,
                                                                          tag='test'))
        self.assertEqual(len(tables), 2)

    def test_get_dbt_state_candidate_view_profile_tag_and_run_results(self):
        tables = dbtutil.get_dbt_state_candidate(self.dbt_state_dir, dict(view_profile=True,
                                                                          dbt_resources=None,
                                                                          tag='piperider'))
        self.assertEqual(len(tables), 4)
        self.assertEqual(tables[0].get('name'), 'PRICE_PRESENT')

    def test_get_dbt_state_tests_result(self):
        results = dbtutil.get_dbt_state_tests_result(self.dbt_state_dir)

        table_names = [r.get('table') for r in results]
        column_names = [r.get('column') for r in results]

        self.assertIn('PRICE_PRESENT', table_names)
        self.assertIn('MA60', column_names)

        status = None
        for r in results:
            if r.get('table') == 'PRICE_PRESENT' and r.get('column') == 'MA60':
                status = r.get('status')
                break
        self.assertEqual('failed', status)

    def test_dbt_append_descriptions(self):
        ds = SqliteDataSource("test")
        engine = ds.get_engine_by_database()
        profiler = Profiler(ds)

        data1 = [
            ("symbol", "open", "ma60", "ma20"),
            ("AAA", 1, 2, 3),
            ("BBB", 4, 5, 6),
        ]
        data2 = [
            ("symbol", "open", "ma60", "ma20"),
            ("CCC", 1, 2, 3),
            ("DDD", 4, 5, 6),
        ]
        create_table(engine, "PRICE", data1)
        create_table(engine, "PRICE_PRESENT", data2)
        profile_results = profiler.profile()
        dbtutil.append_descriptions(profile_results, self.dbt_state_dir)

        self.assertEqual('The price data to today',
                         profile_results['tables']['PRICE_PRESENT']['description'])
        self.assertEqual('The symbol name',
                         profile_results['tables']['PRICE_PRESENT']['columns']['symbol']['description'])

    @skip("deprecated after v0.33.0, skipping")
    def test_get_dbt_state_metrics_only_tag(self):
        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'piperider', None)

        self.assertEqual(len(metrics), 3)

        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'abc', None)

        self.assertEqual(len(metrics), 1)
        self.assertEqual(metrics[0].name, 'active_projects_per_user')

    @skip("deprecated after v0.33.0, skipping")
    def test_get_dbt_state_metrics_only_resources(self):
        metrics = ['metric.infusetude.active_projects',
                   'metric.infusetude.active_projects_per_user',
                   'metric.infusetude.active_users',
                   'metric.infusetude.total_events',
                   'metric.infusetude.total_users']
        resources = dict(models=[], metrics=metrics)
        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'piperider', resources)
        self.assertEqual(len(metrics), 5)

        metrics = dbtutil.get_dbt_state_metrics(self.dbt_state_dir, 'abc', resources)
        self.assertEqual(len(metrics), 5)
        self.assertEqual(metrics[0].name, 'total_events')

    @pytest.mark.skipif(dbt_version < '1.6', reason="only for dbt 1.6")
    @mock.patch('piperider_cli.dbtutil._get_state_manifest')
    def test_get_dbt_state_metrics_16(self, _get_state_manifest):
        _get_state_manifest.return_value = _load_manifest('dbt-duckdb-1.6.0-manifest.json')
        metrics = dbtutil.get_dbt_state_metrics_16(self.dbt_state_dir, dbt_tag=None, dbt_resources=None)

        self.assertEqual(len(metrics), 4)
        # expenses
        self.assertEqual(metrics[0].name, 'expenses')

        # revenue
        self.assertEqual(metrics[1].calculation_method, 'sum')

        # average_order_amount
        self.assertEqual(metrics[2].model.table, 'orders')
        self.assertEqual(metrics[2].model.timestamp, 'order_date')
        self.assertEqual(metrics[2].model.expression, 'amount')

        # profit
        self.assertEqual(metrics[3].model, None)
        self.assertEqual(metrics[3].calculation_method, 'derived')
        self.assertEqual(metrics[3].expression, 'revenue - expenses')

        # skip metric < 1.6
        _get_state_manifest.return_value = _load_manifest('dbt-duckdb-1.5.1-manifest.json')
        metrics = dbtutil.get_dbt_state_metrics_16(self.dbt_state_dir, dbt_tag=None, dbt_resources=None)
        self.assertEqual(len(metrics), 0)

    @mock.patch('pathlib.Path.cwd',
                return_value=Path(os.path.join(os.path.dirname(__file__), 'mock_dbt_project', 'dir_1', 'dir_2')))
    def test_search_dbt_project_path(self, *args):
        path = dbtutil.search_dbt_project_path()
        self.assertEqual(path, os.path.join(os.path.dirname(__file__), 'mock_dbt_project'))

    def test_is_ready(self):
        self.assertFalse(dbtutil.is_ready({
            'profile': None
        }))

        self.assertFalse(dbtutil.is_ready({
            'profile': 'mock_data',
            'target': None
        }))

        self.assertFalse(dbtutil.is_ready({
            'profile': 'mock_data',
            'target': 'mock_data',
            'projectDir': None
        }))

        self.assertTrue(dbtutil.is_ready({
            'profile': 'mock_data',
            'target': 'mock_data',
            'projectDir': '~/',  # this is a valid path
        }))

    @mock.patch('piperider_cli.dbtutil.get_dbt_manifest')
    def test_load_dbt_resources(self, get_dbt_manifest):
        v = dbt_version
        target_path = os.path.join(os.path.dirname(__file__), 'mock_dbt_data')
        if v == '1.6':
            get_dbt_manifest.return_value = _load_manifest('dbt-duckdb-1.6.0-manifest.json')
        elif v == '1.5':
            get_dbt_manifest.return_value = _load_manifest('dbt-duckdb-1.5.1-manifest.json')
        elif v == '1.4':
            get_dbt_manifest.return_value = _load_manifest('dbt-duckdb-1.4.2-manifest.json')
        elif v == '1.3':
            get_dbt_manifest.return_value = _load_manifest('dbt-postgres-1.3.4-manifest.json')
        else:
            raise Exception(f'Unsupported dbt version: {v}')
        resources = dbtutil.load_dbt_resources(target_path)
        self.assertIn('models', resources)
        self.assertIn('metrics', resources)

    def test_get_support_time_grain(self):
        time_grains = dbtutil.get_support_time_grains('day')
        self.assertListEqual(time_grains, ['day', 'month', 'year'])

        time_grains = dbtutil.get_support_time_grains('week')
        self.assertListEqual(time_grains, ['month', 'year'])

        time_grains = dbtutil.get_support_time_grains('month')
        self.assertListEqual(time_grains, ['month', 'year'])

        time_grains = dbtutil.get_support_time_grains('quarter')
        self.assertListEqual(time_grains, ['year'])

        time_grains = dbtutil.get_support_time_grains('year')
        self.assertListEqual(time_grains, ['year'])



import json
import os
import unittest
from typing import Dict, List
from unittest import TestCase

import pytest
from packaging import version

from piperider_cli.dbt import dbt_version
from piperider_cli.dbt.list_task import (
    compare_models_between_manifests,
    list_resources_unique_id_from_manifest,
    load_manifest,
)
from piperider_cli.dbt.changeset import GraphDataChangeSet
from tests.test_dbt_manifest_compatible import _load_manifest


class _BaseDbtTest(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.dbt_state_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data"
        )

    def fake_data_path(self, filename: str):
        return os.path.join(self.dbt_state_dir, filename)

    def manifest_object(self, manifest_name, embedded_from_run=False):
        with open(self.fake_data_path(manifest_name)) as fh:
            data = json.loads(fh.read())
            if embedded_from_run:
                data = data.get("dbt", {}).get("manifest", {})
            return load_manifest(data)

    def run_object(self, run_name):
        with open(self.fake_data_path(run_name)) as fh:
            data = json.loads(fh.read())
            return data

    def load_json(self, filename: str) -> Dict:
        with open(self.fake_data_path(filename)) as fh:
            return json.loads(fh.read())

    def base_manifest(self):
        return self.manifest_object("jaffle_shop_base_1.3.json", True)

    def target_manifest(self):
        return self.manifest_object("jaffle_shop_target_1.3.json", True)

    def base_run(self):
        return self.run_object("jaffle_shop_base_1.3.json")

    def target_run(self):
        return self.run_object("jaffle_shop_target_1.3.json")

    def base_31587(self):
        return self.run_object("sc-31587-base.json")

    def target_31587(self):
        return self.run_object("sc-31587-input.json")

    def base_31587_with_ref(self):
        return self.run_object("sc-31587-with-ref-base.json")

    def target_31587_with_ref(self):
        return self.run_object("sc-31587-with-ref-input.json")

    def base_31587_metrics(self):
        return self.run_object("sc-31587-metrics-base.json")

    def target_31587_metrics(self):
        return self.run_object("sc-31587-metrics-input.json")

    def base_31782(self):
        return self.run_object("sc-31782-base.json")

    def target_31782(self):
        return self.run_object("sc-31782-target.json")


class TestDbtIntegration(_BaseDbtTest):
    def setUp(self):
        super().setUp()

    def assertDbtResources(self, r1: List[str], r2: List[str]):
        if dbt_version >= 'v1.6':
            r1 = [x for x in r1 if not x.startswith('metric')]
            r2 = [x for x in r2 if not x.startswith('metric')]
        return self.assertEqual(r1, r2)

    def test_list_all_resources(self):
        expected = [
            "metric.jaffle_shop.revenue",
            "model.jaffle_shop.customers",
            "model.jaffle_shop.orders",
            "model.jaffle_shop.stg_customers",
            "model.jaffle_shop.stg_orders",
            "model.jaffle_shop.stg_payments",
            "seed.jaffle_shop.raw_customers",
            "seed.jaffle_shop.raw_orders",
            "seed.jaffle_shop.raw_payments",
            "test.jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned.be6b5b5ec3",
            "test.jaffle_shop.accepted_values_stg_orders_status__placed__shipped__completed__return_pending__returned.080fb20aad",
            "test.jaffle_shop.accepted_values_stg_payments_payment_method__credit_card__coupon__bank_transfer__gift_card.3c3820f278",
            "test.jaffle_shop.not_null_customers_customer_id.5c9bf9911d",
            "test.jaffle_shop.not_null_orders_amount.106140f9fd",
            "test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49",
            "test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625",
            "test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59",
            "test.jaffle_shop.not_null_orders_customer_id.c5f02694af",
            "test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a",
            "test.jaffle_shop.not_null_orders_order_id.cf6c17daed",
            "test.jaffle_shop.not_null_stg_customers_customer_id.e2cfb1f9aa",
            "test.jaffle_shop.not_null_stg_orders_order_id.81cfe2fe64",
            "test.jaffle_shop.not_null_stg_payments_payment_id.c19cc50075",
            "test.jaffle_shop.relationships_orders_customer_id__customer_id__ref_customers_.c6ec7f58f2",
            "test.jaffle_shop.unique_customers_customer_id.c5af1ff4b1",
            "test.jaffle_shop.unique_orders_order_id.fed79b3a6e",
            "test.jaffle_shop.unique_stg_customers_customer_id.c7614daada",
            "test.jaffle_shop.unique_stg_orders_order_id.e3b841c71a",
            "test.jaffle_shop.unique_stg_payments_payment_id.3744510712",
        ]

        all_results = list_resources_unique_id_from_manifest(self.base_manifest())

        self.assertDbtResources(expected, all_results)

    @pytest.mark.skipif(dbt_version < 'v1.6', reason='skip manifest test before dbt-core 1.6')
    def test_list_all_resources_16(self):
        expected = [
            "metric.jaffle_shop.average_order_amount",
            "metric.jaffle_shop.expenses",
            "metric.jaffle_shop.profit",
            "metric.jaffle_shop.revenue",
            "model.jaffle_shop.int_customer_order_history_joined",
            "model.jaffle_shop.int_order_payments_pivoted",
            "model.jaffle_shop.metricflow_time_spine",
            "model.jaffle_shop.orders",
            "model.jaffle_shop.stg_customers",
            "model.jaffle_shop.stg_orders",
            "model.jaffle_shop.stg_payments",
            "seed.jaffle_shop.raw_customers",
            "seed.jaffle_shop.raw_orders",
            "seed.jaffle_shop.raw_payments",
            "test.jaffle_shop.accepted_values_int_order_payments_pivoted_status__placed__shipped__completed__return_pending__returned.0ccdff53e8",
            "test.jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned.be6b5b5ec3",
            "test.jaffle_shop.accepted_values_stg_orders_status__placed__shipped__completed__return_pending__returned.080fb20aad",
            "test.jaffle_shop.accepted_values_stg_payments_payment_method__credit_card__coupon__bank_transfer__gift_card.3c3820f278",
            "test.jaffle_shop.not_null_int_customer_order_history_joined_customer_id.5eeb8cdf92",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_amount.b7598e0e3b",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_bank_transfer_amount.1a9e62933b",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_coupon_amount.2532b538c2",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_credit_card_amount.ae9c42d967",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_customer_id.3db59c6de4",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_gift_card_amount.710d789cc0",
            "test.jaffle_shop.not_null_int_order_payments_pivoted_order_id.787ba994a8",
            "test.jaffle_shop.not_null_orders_amount.106140f9fd",
            "test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49",
            "test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625",
            "test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59",
            "test.jaffle_shop.not_null_orders_customer_id.c5f02694af",
            "test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a",
            "test.jaffle_shop.not_null_orders_order_id.cf6c17daed",
            "test.jaffle_shop.not_null_stg_customers_customer_id.e2cfb1f9aa",
            "test.jaffle_shop.not_null_stg_orders_order_id.81cfe2fe64",
            "test.jaffle_shop.not_null_stg_payments_payment_id.c19cc50075",
            "test.jaffle_shop.relationships_int_order_payments_pivoted_customer_id__customer_id__ref_int_customer_order_history_joined_.654a1aa35d",
            "test.jaffle_shop.unique_int_customer_order_history_joined_customer_id.995635f7d9",
            "test.jaffle_shop.unique_int_order_payments_pivoted_order_id.34a0f3307d",
            "test.jaffle_shop.unique_orders_order_id.fed79b3a6e",
            "test.jaffle_shop.unique_stg_customers_customer_id.c7614daada",
            "test.jaffle_shop.unique_stg_orders_order_id.e3b841c71a",
            "test.jaffle_shop.unique_stg_payments_payment_id.3744510712",
        ]

        manifest = load_manifest(_load_manifest('dbt-duckdb-1.6.0-manifest.json'))
        all_results = list_resources_unique_id_from_manifest(manifest)

        self.assertDbtResources(expected, all_results)

    def test_compare_with_manifests(self):
        without_downstream = compare_models_between_manifests(
            self.base_manifest(), self.target_manifest()
        )

        expected = ["jaffle_shop.customers", "jaffle_shop.orders"]
        self.assertDbtResources(without_downstream, expected)

        with_downstream = compare_models_between_manifests(
            self.base_manifest(), self.target_manifest(), True
        )

        expected = ["jaffle_shop.customers", "jaffle_shop.orders"]
        self.assertDbtResources(with_downstream, expected)

    def test_list_explicit_changes(self):
        c = GraphDataChangeSet(self.base_run(), self.target_run())

        expected = ["model.jaffle_shop.customers", "model.jaffle_shop.orders"]
        changes = c.list_explicit_changes()
        self.assertDbtResources(changes, expected)

        print(c.list_implicit_changes())

    @unittest.skipIf(
        dbt_version < version.parse("1.4"),
        "this only works after manifests generated after the v1.4",
    )
    def test_list_explicit_changes_without_ref_ids(self):
        c = GraphDataChangeSet(self.base_31587(), self.target_31587())

        expected = ["model.jaffle_shop.orders"]
        changes = c.list_explicit_changes()
        self.assertDbtResources(changes, expected)

        expected_implicit = [
            "metric.jaffle_shop.average_order_amount",
            "model.jaffle_shop.orders",
        ]
        self.assertDbtResources(c.list_implicit_changes(), expected_implicit)

    @unittest.skipIf(
        dbt_version < version.parse("1.4"),
        "this only works after manifests generated after the v1.4",
    )
    def test_list_explicit_changes_with_ref_ids(self):
        c = GraphDataChangeSet(self.base_31587_with_ref(), self.target_31587_with_ref())

        expected = ["model.jaffle_shop.orders"]
        changes = c.list_explicit_changes()
        self.assertDbtResources(changes, expected)

        expected_implicit = [
            "metric.jaffle_shop.average_order_amount",
            "model.jaffle_shop.orders",
        ]
        self.assertDbtResources(c.list_implicit_changes(), expected_implicit)

    @unittest.skipIf(
        dbt_version < version.parse("1.4"),
        "this only works after manifests generated after the v1.4",
    )
    def test_list_changes_metrics_case(self):
        c = GraphDataChangeSet(self.base_31587_metrics(), self.target_31587_metrics())
        expected = ["model.jaffle_shop.orders"]
        changes = c.list_explicit_changes()
        self.assertDbtResources(changes, expected)

        changes = c.list_implicit_changes()
        self.assertDbtResources(
            changes,
            ["metric.jaffle_shop.average_order_amount", "model.jaffle_shop.orders"],
        )

    @unittest.skipIf(
        dbt_version < version.parse("1.4"),
        "this only works after manifests generated after the v1.4",
    )
    def test_list_large_seeds_case(self):
        try:
            c = GraphDataChangeSet(self.base_31782(), self.target_31782())
        except Exception as e:
            self.fail("Unexpected Exception")

        expected = ["metric.jaffle_shop.expenses", "model.jaffle_shop.stg_payments"]

        changes = c.list_explicit_changes()
        self.assertDbtResources(changes, expected)

        expected = [
            "metric.jaffle_shop.expenses",
            "metric.jaffle_shop.profit",
            "model.jaffle_shop.stg_payments",
        ]

        changes = c.list_implicit_changes()
        self.assertDbtResources(changes, expected)



import os
import warnings

from sqlalchemy.exc import SAWarning

from . import DataSource
from .field import PathField


class SqliteDataSource(DataSource):

    def __init__(self, name, **kwargs):
        super().__init__(name, 'sqlite', **kwargs)

        # in dbt case, we should push the dbpath back to the credential
        if 'dbt' in kwargs:
            dbpath = kwargs.get('credential', {}).get('schemas_and_paths', {}).get('main')
            kwargs.get('credential', {})['dbpath'] = dbpath
        self.credential_source = 'config'
        self.fields = [
            PathField('dbpath', description='Path of database file'),
        ]
        warnings.filterwarnings('ignore',
                                r'^Dialect sqlite\+pysqlite does \*not\* support Decimal objects natively.*$',
                                SAWarning)

    def validate(self):
        if self.type_name != 'sqlite':
            raise ValueError('type name should be sqlite')
        return self._validate_required_fields()

    def to_database_url(self, database):
        credential = self.credential
        from piperider_cli.configuration import FileSystem
        dbpath = credential.get('dbpath')
        if dbpath is None:
            return "sqlite://"
        else:
            if os.path.isabs(dbpath) is False:
                dbpath = os.path.join(FileSystem.WORKING_DIRECTORY, dbpath)
            sqlite_file = os.path.abspath(dbpath)
            if not os.path.exists(sqlite_file):
                raise ValueError(f'Cannot find the sqlite at {sqlite_file}')
            return f"sqlite:///{sqlite_file}"

    def engine_args(self):
        return {
            'isolation_level': 'AUTOCOMMIT',
        }

    def verify_connector(self):
        # sqlite is builtin connector
        return None

    def _get_display_description(self):
        return f"type={self.type_name}, dbpath={self.credential.get('dbpath')}"

    def get_database(self):
        return self.credential.get('database')

    def get_schema(self):
        return self.credential.get('schema')



from io import StringIO
from unittest import TestCase

from ruamel import yaml

from piperider_cli.assertion_engine import AssertionEngine


def _(assertion_content: str):
    def content_provider(self):
        self.assertions_content = yaml.safe_load(StringIO(assertion_content))
        return [], []

    return content_provider


class BuiltinValidationTests(TestCase):

    def setUp(self) -> None:
        self.engine = AssertionEngine(None)
        self.origin_function = AssertionEngine.load_assertion_content

    def tearDown(self) -> None:
        AssertionEngine.load_assertion_content = self.origin_function

    def test_validation_assert_row_count_in_range(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count_in_range
                assert:
                  count: [1000, 200000]
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

    def test_validation_assert_row_count_in_range_no_args(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count_in_range
                assert:
                  # it should be "count"
                  range: [1000, 200000]
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_row_count_in_range for table orders_1k
ERROR: count parameter is required
ERROR: count parameter should be a list""", results[0].as_internal_report())

    def test_validation_assert_row_count_in_range_invalid_args(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count_in_range
                assert:
                  count: [1000, 200000, 2]
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_row_count_in_range for table orders_1k
ERROR: count parameter should contain two values""", results[0].as_internal_report())

    def test_validation_assert_row_count(self):
        # test with valid format with min only
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  min: 10
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

        # test with valid format with max only
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  max: 10
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

        # test with valid format with min and max
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  min: 10
                  max: 100
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

    def test_validation_assert_row_count_invalid_args(self):
        # test with invalid syntax
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  x: 10
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))
        self.assertEqual("""Found assertion syntax problem => name: assert_row_count for table orders_1k
ERROR: There should contain any parameter names in ['min', 'max']""", results[0].as_internal_report())

        # test with invalid syntax: max < min
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  min: 100
                  max: 1
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_row_count for table orders_1k
ERROR: The max value should be greater than or equal to the min value.""", results[0].as_internal_report())

        # test with invalid syntax: not int
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: assert_row_count
                assert:
                  min: 1.0
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_row_count for table orders_1k
ERROR: min parameter should be a int value""", results[0].as_internal_report())

    def test_validation_assert_column_not_null(self):
        self._test_no_args_assertion('assert_column_not_null')

    def test_validation_assert_column_null(self):
        self._test_no_args_assertion('assert_column_null')

    def test_validation_assert_column_unique(self):
        self._test_no_args_assertion('assert_column_unique')

    def test_validation_assert_column_exist(self):
        self._test_no_args_assertion('assert_column_exist')

    def _test_no_args_assertion(self, function_name):
        # test with no problem syntax
        AssertionEngine.load_assertion_content = _(f"""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: {function_name}
            """)
        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)
        # test with error syntax
        AssertionEngine.load_assertion_content = _(f"""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: {function_name}
                    assert:
                      foo: bar
            """)
        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))
        self.assertEqual(f"""Found assertion syntax problem => name: {function_name} for table orders_1k and column foobarbar
ERROR: parameters are not allowed""", results[0].as_internal_report())

    def test_validation_assert_column_min_in_range(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: assert_column_min_in_range
                    assert:
                        min: [10, 30]
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

    def test_validation_assert_column_min_in_range_date(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: assert_column_min_in_range
                    assert:
                        min: [2022-05-20, 2022-05-31]
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertListEqual([], results)

    def test_validation_assert_column_min_in_range_no_args(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: assert_column_min_in_range
                    assert:
                        # it should be "min"
                        max: [10, 100]
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_column_min_in_range for table orders_1k and column foobarbar
ERROR: min parameter is required""", results[0].as_internal_report())

    def test_validation_assert_column_min_in_range_invalid_args(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              columns:
                foobarbar:
                  tests:
                  - name: assert_column_min_in_range
                    assert:
                        min: [10, 2022-05-23]
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))

        self.assertEqual("""Found assertion syntax problem => name: assert_column_min_in_range for table orders_1k and column foobarbar
ERROR: min parameter should be the same types""", results[0].as_internal_report())

    def test_validation_assert_no_such_assertion(self):
        # test with valid format
        AssertionEngine.load_assertion_content = _("""
            orders_1k:  # Table Name
              # Test Cases for Table
              tests:
              - name: there_is_no_such_assertion
                assert:
                  count: [1000, 200000]
                tags:
                - OPTIONAL
            """)

        # expect no errors and warnings
        self.engine.load_all_assertions_for_validation()
        results = self.engine.validate_assertions()
        self.assertEqual(1, len(results))
        self.assertEqual("""Found assertion syntax problem => name: there_is_no_such_assertion for table orders_1k
ERROR: cannot find an assertion by name 'there_is_no_such_assertion'""",
                         results[0].as_internal_report())



import os
import shutil
from unittest import TestCase

from piperider_cli.configuration import FileSystem, Configuration
from piperider_cli.error import PipeRiderConnectorUnsupportedError
from piperider_cli.profiler import ProfileSubject
from piperider_cli.runner import _filter_subject, get_dbt_profile_subjects, Runner
from piperider_cli.statistics import Statistics


class TestRunnerSubjectFilter(TestCase):

    def setUp(self):
        self.dbt_state_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data")
        self.statistics = Statistics()
        self.statistics.reset()

    def test_filter_subject(self):
        self.assertTrue(_filter_subject('a', None, None))
        self.assertTrue(_filter_subject('a', ['a', 'b'], None))
        self.assertFalse(_filter_subject('a', [], None))
        self.assertFalse(_filter_subject('a', None, ['a', 'b']))
        self.assertTrue(_filter_subject('a', None, []))
        self.assertFalse(_filter_subject('a', ['a'], ['a']))
        self.assertTrue(_filter_subject('a', ['a'], ['b']))
        self.assertTrue(_filter_subject('a', ['A'], ['B']))

    def test_dbt_profile_subjects_tags_and_config(self):
        options = dict(
            view_profile=False,
            dbt_resources=None,
            tag='test'
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['project'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(2, len(subjects))
        self.assertEqual('stg_event', subjects[0].name)
        self.assertEqual('session', subjects[1].name)

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(11, self.statistics.statistic['notag'])

    def test_dbt_profile_subjects_dbt_resources_and_config(self):
        options = dict(
            view_profile=None,
            dbt_resources=dict(
                models=['model.infusetude.project'],
                metrics=[]
            ),
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, None, ['project'])

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('project', subjects[0].name)

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(12, self.statistics.statistic['filter'])

    def test_dbt_profile_subjects_config_excludes(self):
        options = dict(
            view_profile=None,
            dbt_resources=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, None, ['session'])

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(8, len(subjects))
        self.assertNotIn('session', [s.name for s in subjects])

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(1, self.statistics.statistic['filter'])
        self.assertEqual(1, self.statistics.statistic['view'])

    def test_dbt_profile_subjects_config_includes(self):
        options = dict(
            view_profile=None,
            dbt_resources=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['user'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('user', subjects[0].name)

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(8, self.statistics.statistic['filter'])
        self.assertEqual(1, self.statistics.statistic['view'])

    def test_dbt_profile_subjects_config_view_includes(self):
        options = dict(
            view_profile=True,
            dbt_resources=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['stg_event'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('stg_event', subjects[0].name)

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(9, self.statistics.statistic['filter'])

    def test_dbt_profile_subjects_config_no_view_includes(self):
        options = dict(
            view_profile=False,
            dbt_resources=None,
            tag=None
        )

        def filter_fn(subject: ProfileSubject):
            return _filter_subject(subject.name, ['stg_event', 'user'], None)

        subjects = get_dbt_profile_subjects(self.dbt_state_dir, options, filter_fn)

        self.assertEqual(1, len(subjects))
        self.assertEqual('user', subjects[0].name)

        self.assertEqual(10, self.statistics.statistic['total'])
        self.assertEqual(8, self.statistics.statistic['filter'])
        self.assertEqual(1, self.statistics.statistic['view'])


class TestExecuteRunnerWithoutDataSourceConnection(TestCase):
    def setUp(self) -> None:
        self.original_working_directory = FileSystem.WORKING_DIRECTORY
        self.original_profiles_dir = FileSystem.dbt_profiles_dir
        self.dbt_project_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mock_dbt_project',
                                            'skip_datasource_connection')

        FileSystem.set_working_directory(self.dbt_project_dir)
        FileSystem.set_dbt_profiles_dir(self.dbt_project_dir)
        Configuration.cleanup()
        pass

    def test_skip_datasource(self):
        rc = Runner.exec(skip_datasource_connection=True)
        assert rc == 0

    def test_non_skip_datasource(self):
        # PipeRider has not supported 'hive' datasource yet
        with self.assertRaises(PipeRiderConnectorUnsupportedError):
            Runner.exec(skip_datasource_connection=False)

    def tearDown(self) -> None:
        FileSystem.set_working_directory(self.original_working_directory)
        FileSystem.dbt_profiles_dir = self.original_profiles_dir

        # Clean up piperider output files
        piperider_output_dir = os.path.join(self.dbt_project_dir, '.piperider', 'outputs')
        try:
            shutil.rmtree(piperider_output_dir)
        except FileNotFoundError:
            pass
        Configuration.cleanup()
        pass



import abc
import os
import re
import shlex
import subprocess
import tarfile
import tempfile
from pathlib import Path
from subprocess import Popen
from typing import Dict, Tuple

from rich.console import Console

from piperider_cli.configuration import FileSystem
from piperider_cli.dbt.list_task import list_resources_unique_id_from_manifest, load_full_manifest, load_manifest
from piperider_cli.dbtutil import get_dbt_manifest
from piperider_cli.error import PipeRiderError, RecipeException


class AbstractRecipeUtils(metaclass=abc.ABCMeta):
    def __init__(self, console: Console):
        self.console = console

    @abc.abstractmethod
    def execute_command_in_silent(
            self, command_line, env: Dict = None
    ) -> Tuple[str, str, int]:
        """
        Execute command without showing outputs

        the outputs will be collected to the method returns

        :returns: stdout, stderr, exit_code
        """

    @abc.abstractmethod
    def execute_command_with_showing_output(
            self, command_line, env: Dict = None
    ) -> int:
        """
        Execute command and showing outputs

        the outputs will be showing to the terminal

        :returns: exit_code
        """

    @staticmethod
    def dryrun_ignored_execute_command(command_line, env: Dict = None):
        cmd = shlex.split(command_line)
        proc = None
        try:
            proc = Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env or os.environ.copy(),
                cwd=FileSystem.WORKING_DIRECTORY,
            )
            outs, errs = proc.communicate()
        except BaseException as e:
            if proc:
                proc.kill()
                outs, errs = proc.communicate()
            else:
                return None, e, 1

        if outs is not None:
            outs = outs.decode().strip()
        if errs is not None:
            errs = errs.decode().strip()
        return outs, errs, proc.returncode

    @staticmethod
    def dryrun_ignored_execute_command_no_outputs(command_line, env: Dict = None):
        cmd = shlex.split(command_line)
        proc = None

        try:
            proc = Popen(cmd, env=env or os.environ.copy(), cwd=FileSystem.WORKING_DIRECTORY)
            proc.communicate()
        except BaseException:
            if proc:
                proc.kill()
                proc.communicate()
            else:
                return 1

        return proc.returncode

    def git_branch(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            "git rev-parse --abbrev-ref HEAD"
        )
        if exit_code != 0:
            return None

        return outs

    def git_checkout_to(self, commit_or_branch):
        outs, errs, exit_code = self.execute_command_in_silent(
            f"git checkout {commit_or_branch}"
        )
        if exit_code != 0:
            raise RecipeException(errs)

    def git_merge_base(self, a: str, b: str):
        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            f"git merge-base {a} {b}"
        )
        if exit_code != 0:
            ex = RecipeException(errs)
            if outs == "" and errs == "":
                ex.message = "Empty result from git merge-base"
                ex.hint = 'Please try "git fetch --unshallow" first'
            elif re.match(r"fatal: Not a valid object name (.*)", errs):
                matched = re.match(r"fatal: Not a valid object name (.*)", errs)
                ex.message = f"Invalid git branch: {matched.group(1)}"
                ex.hint = (
                    f"Please check is the branch name '{matched.group(1)}' correct?"
                )
            raise ex
        return outs

    def git_archive(self, commit_or_branch):
        def untar(file_path, extract_dir):
            with tarfile.open(file_path, 'r') as tar:
                tar.extractall(extract_dir)

        tmpdirname = Path(tempfile.mkdtemp())
        tar = (tmpdirname / f'{commit_or_branch}.tar').as_posix()

        outs, errs, exit_code = self.execute_command_in_silent(
            rf'git archive --format=tar --output={tar} {commit_or_branch}'
        )
        print(rf'git archive --format=tar --output={tar} {commit_or_branch}')
        if exit_code != 0:
            raise RecipeException(errs)

        project_dir = (tmpdirname / commit_or_branch).as_posix()
        untar(tar, project_dir)

        return project_dir

    def ensure_git_ready(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command("git --version")
        if exit_code != 0:
            raise PipeRiderError("git is not installed", hint="Please install it first")

        if "version" not in outs:
            raise RecipeException("Unknown response from git --version")

        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            "git status --porcelain"
        )
        if exit_code != 0 and "not a git repository" in errs:
            raise RecipeException("The working directory is not a git repository.")

    def check_dbt_command(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command("dbt --version")
        if exit_code != 0:
            raise PipeRiderError("dbt is not installed", hint="Please install it first")

    def list_dbt_resources(self, target_path, select=None, state=None):
        if state:
            manifest = load_full_manifest(target_path)
        else:
            manifest = load_manifest(get_dbt_manifest(target_path))
        return list_resources_unique_id_from_manifest(manifest, select=select, state=state)


class RecipeUtils(AbstractRecipeUtils):
    def execute_command_with_showing_output(self, command_line, env: Dict = None):
        return self.dryrun_ignored_execute_command_no_outputs(command_line, env)

    def execute_command_in_silent(self, command_line, env: Dict = None):
        print("execute_command_in_silent", command_line)
        outs, errs, exit_code = self.dryrun_ignored_execute_command(command_line, env)
        return outs, errs, exit_code


class DryRunRecipeUtils(AbstractRecipeUtils):
    def execute_command_in_silent(self, command_line, env: Dict = None):
        self.console.print(f"[green]:external-command:>[/green] {command_line}")
        return "", "", 0

    def execute_command_with_showing_output(self, command_line, env: Dict = None):
        self.console.print(f"[green]:external-command:>[/green] {command_line}")
        return 0

    def git_archive(self, commit_or_branch):
        cmd = f"git archive --format=tar --output=/path/to/tmp/{commit_or_branch}.tar {commit_or_branch}"
        self.console.print(f"[green]:external-command:>[/green] [default]{cmd}[/default]")
        return '/path/to/tmp'

    def list_dbt_resources(self, target_path, select=None, state=None):
        state_msg = ''
        if state:
            state_msg = f'state: {state}'
        self.console.print(f"[green]:dbt-resources:>[/green] with select: {''.join(select)} {state_msg}")
        return ['dbt_model1', 'dbt_model2', 'dbt_model3']

    def remove_dir(self, path):
        self.console.print(f"[green]:remove-dir:>[/green] {path}")


class InteractiveStopException(Exception):
    def __init__(self):
        pass


class InteractiveRecipeDecorator(AbstractRecipeUtils):
    def __init__(self, utils: AbstractRecipeUtils):
        self.decoratee = utils

    def execute_command_in_silent(
            self, command_line, env: Dict = None
    ) -> Tuple[str, str, int]:
        # ask user continue
        if self.should_continue(command_line):
            outs, errs, exit_code = self.decoratee.execute_command_in_silent(
                command_line, env
            )
            return outs, errs, exit_code
        raise InteractiveStopException()

    def execute_command_with_showing_output(
            self, command_line, env: Dict = None
    ) -> int:
        # ask user continue
        if self.should_continue(command_line):
            return self.decoratee.execute_command_with_showing_output(command_line, env)
        raise InteractiveStopException()

    @staticmethod
    def should_continue(command_line: str):
        try:
            return InteractiveRecipeDecorator.ask_for_confirmation(command_line)
        except KeyboardInterrupt:
            raise InteractiveStopException()

    @staticmethod
    def ask_for_confirmation(command_line: str):
        from piperider_cli.datasource import FANCY_USER_INPUT

        if FANCY_USER_INPUT:
            import inquirer
            return inquirer.confirm(f"Execute '{command_line}'", default=True)
        else:
            from rich.prompt import Confirm
            return Confirm.ask(f"Execute '{command_line}'")

    def git_archive(self, commit_or_branch) -> str:
        # ask user continue
        cmd = f"git archive --format=tar --output=/path/to/tmp/{commit_or_branch}.tar {commit_or_branch}"
        if self.should_continue(cmd):
            return self.decoratee.git_archive(commit_or_branch)
        raise InteractiveStopException()


if __name__ == "__main__":
    utils = RecipeUtils(Console())
    utils = DryRunRecipeUtils(Console())
    utils.ensure_git_ready()
    utils.git_checkout_to("abc")
